#include "magic_probability_toolkit.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

void article_01()
{
	const int prefered_lower_bound = 3;
	const int prefered_upper_bound = 5;

	int hand_sizes[] = { 2, 7 };
	for ( int idx = 0; idx < std::extent< decltype( hand_sizes ) >::value; ++idx )
	{
		CProbabilityDistribution state( 40, 17 );
		state.DrawCards( hand_sizes[ idx ] );

		CTable table( "Land count distribution in a %d-card hand.", hand_sizes[ idx  ] );
		table.AddColumn( "land" );
		table.AddColumn( "probability" );

		for ( auto i = state.m_distribution.begin(); i != state.m_distribution.end(); ++i )
		{
			const CProbabilityDistributionEntry& entry = *i;
			table.AddInteger( entry.m_player.m_hand.m_lands );
			table.AddProbability( entry.m_probability );
		}

		const double screwed_probability = state.TakeMeasurement(
			[ prefered_lower_bound, prefered_upper_bound ]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands < prefered_lower_bound;
			}
		);
		const double preferred_probability = state.TakeMeasurement(
			[ prefered_lower_bound, prefered_upper_bound ]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands >= prefered_lower_bound && entry.m_player.m_hand.m_lands <= prefered_upper_bound;
			}
		);
		const double flooded_probability = state.TakeMeasurement(
			[ prefered_lower_bound, prefered_upper_bound ]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands > prefered_upper_bound;
			}
		);

		table.AddAddendum( "screwed" );
		table.AddProbability( screwed_probability );
		table.AddAddendum( "preferred" );
		table.AddProbability( preferred_probability );
		table.AddAddendum( "flooded" );
		table.AddProbability( flooded_probability );

		table.Print();
	}
}