#include "magic_probability_toolkit.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

void article_04()
{
	const int land_drops[] = { 3, 4 };
	for ( int land_drop_idx = 0; land_drop_idx < std::extent< decltype( land_drops ) >::value; ++land_drop_idx )
	{
		const int land_drop = land_drops[ land_drop_idx ];

		CTable table( "Chance of hitting land drop on turn %d.", land_drop );
		table.AddColumn( "lands in starting hand" );
		table.AddColumn( "probability" );
		const int lands_in_starting_hand[] = { 2, 3 };
		for ( int lands_in_starting_hand_idx = 0; lands_in_starting_hand_idx < std::extent< decltype( lands_in_starting_hand ) >::value; ++lands_in_starting_hand_idx )
		{
			CProbabilityDistribution state( 40, 17, 7, lands_in_starting_hand[ lands_in_starting_hand_idx ] );
			state.DrawCards( land_drop );
			const double probability_of_third_land =  state.TakeMeasurement(
				[ land_drop ]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= land_drop;
				}
			);
			table.AddInteger( lands_in_starting_hand[ lands_in_starting_hand_idx ] );
			table.AddProbability( probability_of_third_land );
		}
		table.Print();
	}

	int converted_casting_costs[] = { 4, 6 };
	for ( int idx = 0; idx < std::extent< decltype( converted_casting_costs ) >::value; ++idx )
	{
		const int starting_hand_size = 7;
		const int converted_casting_cost = converted_casting_costs[ idx ];

		CProbabilityDistribution states[ 7 ];

		for ( int starting_land_count = 0; starting_land_count < std::extent< decltype( states ) >::value; ++starting_land_count )
		{
			CProbabilityDistribution& state = states[ starting_land_count ];

			state = CProbabilityDistribution( 40, 17, starting_hand_size, starting_land_count );
			state.DrawUntil(
				[ starting_hand_size, converted_casting_cost ]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= converted_casting_cost && entry.m_player.m_hand.GetSize() >= starting_hand_size + converted_casting_cost;
				}
			);
		}
		
		{
			CTable table( "Chance of not having %dth land in play on a given turn with a 2 or 3-land starting hand.", converted_casting_cost );
			table.AddColumn( "turn" );
			table.AddColumn( "2-land hand" );
			table.AddColumn( "3-land hand" );
			for ( int turn = 0; turn < 11; ++turn )
			{
				const double probability_two_land = states[ 2 ].TakeMeasurement(
					[ starting_hand_size, turn ]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.GetSize() - starting_hand_size <= turn;
					}
				);
				const double probability_three_land = states[ 3 ].TakeMeasurement(
					[ starting_hand_size, turn ]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.GetSize() - starting_hand_size <= turn;
					}
				);

				if ( probability_two_land > 0.01 || probability_three_land > 0.01 )
				{
					table.AddInteger( turn );
					table.AddProbability( 1.0 - probability_two_land );
					table.AddProbability( 1.0 - probability_three_land );
				}
			}
			table.Print();
		}

		{
			CTable table( "Average turn you'll hit your %dth land drop for a given starting land count.", converted_casting_cost );
			table.AddColumn( "land" );
			table.AddColumn( "average drop turn" );
			assert( converted_casting_cost + 1 <= std::extent< decltype( states ) >::value );
			for ( int starting_land_count = converted_casting_cost + 1; starting_land_count--; )
			{
				const double average_drop_turn = states[ starting_land_count ].TakeMeasurement(
					[ starting_hand_size ]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.GetSize() - starting_hand_size;
					}
				);

				table.AddInteger( starting_land_count );
				table.AddNumber( average_drop_turn );
			}
			table.Print();
		}
	}
}