#include "magic_probability_toolkit.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

void article_05()
{
	// For this calculation I use lands (CardSet::IgnoringColor::Land) to represent the cards I am hoping to draw

	{
		const int turns = 9;

		CTable table( "Chance of drawing a card in the first %d turns with 1 copy in deck.", turns );

		CProbabilityDistribution state( 40, 1 );
		state.DrawCards( 7 );
		state.DrawCards( turns );

		const double chance_of_drawing_card = state.TakeMeasurement(
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands >= 1;
			}
		);

		table.AddAddendum( "probability" );
		table.AddProbability( chance_of_drawing_card );
		table.Print();
	}

	{
		const int turns = 9;

		CTable table( "Chance of drawing a card in the first %d turns with a given number of copies in deck.", turns );
		table.AddColumn( "copies" );
		table.AddColumn( "probability" );

		for ( int copies = 1; copies < 5; ++copies )
		{
			CProbabilityDistribution state( 40, copies );
			state.DrawCards( 7 );
			state.DrawCards( turns );

			const double chance_of_drawing_card = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 1;
				}
			);

			table.AddInteger( copies );
			table.AddProbability( chance_of_drawing_card );
		}

		table.Print();
	}

	{
		const int initial_turn = 6;

		CTable table( "Chance of drawing a card on turn %d with 1 copy in library.", initial_turn );

		CProbabilityDistribution state( 40, 1, 7 + initial_turn - 1, 0 );
		state.DrawCard();

		const double chance_of_drawing_card = state.TakeMeasurement(
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands >= 1;
			}
		);
		table.AddAddendum( "probability" );
		table.AddProbability( chance_of_drawing_card );
		table.Print();
	}

	{
		const int initial_turn = 6;

		CTable table( "Chance of drawing a card on turn %d with a given number of copies in library.", initial_turn );
		table.AddColumn( "copies" );
		table.AddColumn( "probability" );

		for ( int copies = 1; copies < 4; ++copies )
		{
			CProbabilityDistribution state( 40, copies, 7 + initial_turn - 1, 0 );
			state.DrawCard();

			const double chance_of_drawing_card = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 1;
				}
			);

			table.AddInteger( copies );
			table.AddProbability( chance_of_drawing_card );
		}

		table.Print();
	}
	
	{
		const int initial_turn = 6;

		CTable table( "Chance of drawing a card in a given number of turns starting on turn %d with 1 copy in library.", initial_turn );
		table.AddColumn( "turns" );
		table.AddColumn( "probability" );

		for ( int turns = 1; turns < 5; ++turns )
		{
			CProbabilityDistribution state( 40, 1, 7 + initial_turn - 1, 0 );
			state.DrawCards( turns );

			const double chance_of_drawing_card = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 1;
				}
			);

			table.AddInteger( turns );
			table.AddProbability( chance_of_drawing_card );
		}

		table.Print();
	}
		
	{
		const int initial_turn = 6;

		CTable table( "Chance of drawing a card in a given number of turns starting on turn %d with a given number of copies in library.", initial_turn );
		table.AddColumn( "turns" );
		table.AddColumn( "one copy" );
		table.AddColumn( "two copies" );
		table.AddColumn( "three copies" );

		for ( int turns = 1; turns < 5; ++turns )
		{
			table.AddInteger( turns );

			for ( int copies = 1; copies < 4; ++copies )
			{
				CProbabilityDistribution state( 40, copies, 7 + initial_turn - 1, 0 );
				state.DrawCards( turns );

				const double chance_of_drawing_card = state.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.m_lands >= 1;
					}
				);

				table.AddProbability( chance_of_drawing_card );
			}
		}

		table.Print();
	}

	{
		const int starting_life = 10;
		const int next_turn = 6;
		const int damage_per_turn = 2;
		const int previous_draws = next_turn - 1;
		const int draws_before_death = ( starting_life - 1 ) / damage_per_turn;

		CTable table( "I wait until the last possible moment to use an answer in my hand in hopes that I will draw a more efficient answer, of which I have a given number of copies in my library.  What chance do I have of drawing the better answer before I am forced to use the answer in my hand?  What is the average amount of life I lose by waiting?  The threat I am answering is doing %d damage per turn and I have %d life.  It's my opponent's attack phase and my next turn is turn %d.", damage_per_turn, starting_life, next_turn );
		table.AddColumn( "copies" );
		table.AddColumn( "probability" );
		table.AddColumn( "life paid" );

		for ( int copies = 1; copies < 4; ++copies )
		{
			CProbabilityDistribution state( 40, copies, 7 + previous_draws, 0 );
			state.DrawUntil(
				[ previous_draws, draws_before_death ]( const CProbabilityDistributionEntry& entry ) -> bool
				{
					const int draws_so_far = entry.m_player.m_hand.GetSize() - ( 7 + previous_draws );
					return entry.m_player.m_hand.m_lands >= 1 || draws_so_far == draws_before_death;
				}
			);

			const double life_paid = state.TakeMeasurement(
				[ previous_draws, damage_per_turn ]( const CProbabilityDistributionEntry& entry ) -> double
				{
					const int number_of_attacks = entry.m_player.m_hand.GetSize() - ( 7 + previous_draws );
					return damage_per_turn * number_of_attacks;
				}
			);

			const double chance_of_drawing_card = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 1;
				}
			);

			table.AddInteger( copies );
			table.AddProbability( chance_of_drawing_card );
			table.AddNumber( life_paid );
		}

		table.Print();
	}

	{
		const int starting_life = 10;
		const int damage_per_turn = 2;
		const int copies = 2;
		const int draws_before_death = ( starting_life - 1 ) / damage_per_turn;


		CTable table( "I wait until the last possible moment to use an answer in my hand in hopes that I will draw a more efficient answer, of which I have %d copies in my library.  What chance do I have of drawing the better answer before I am forced to use the answer in my hand?  What is the average amount of life I lose by waiting?  The threat I am answering is doing %d damage per turn and I have %d life.  It's my opponent's attack phase and my next turn is the given turn.", copies, damage_per_turn, starting_life );
		table.AddColumn( "next turn" );
		table.AddColumn( "probability" );
		table.AddColumn( "life paid" );

		const int next_turns[] = { 3, 6, 9, 12 };
		for ( int next_turn_idx = 0; next_turn_idx < std::extent< decltype( next_turns ) >::value; ++next_turn_idx )
		{
			const int next_turn = next_turns[ next_turn_idx ];
			const int previous_draws = next_turn - 1;

			CProbabilityDistribution state( 40, copies, 7 + previous_draws, 0 );
			state.DrawUntil(
				[ previous_draws, draws_before_death ]( const CProbabilityDistributionEntry& entry ) -> bool
				{
					const int draws_so_far = entry.m_player.m_hand.GetSize() - ( 7 + previous_draws );
					return entry.m_player.m_hand.m_lands >= 1 || draws_so_far == draws_before_death;
				}
			);

			const double life_paid = state.TakeMeasurement(
				[ previous_draws, damage_per_turn ]( const CProbabilityDistributionEntry& entry ) -> double
				{
					const int number_of_attacks = entry.m_player.m_hand.GetSize() - ( 7 + previous_draws );
					return damage_per_turn * number_of_attacks;
				}
			);

			const double chance_of_drawing_card = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 1;
				}
			);

			table.AddInteger( next_turn );
			table.AddProbability( chance_of_drawing_card );
			table.AddNumber( life_paid );
		}
		
		table.Print();
	}
}