#include "magic_probability_toolkit.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

void article_07()
{
	{
		CTable table( "Probability of casting keyrune on turn three with a two-land starting hand." );
		table.AddColumn( "" );
		table.AddColumn( "probability" );

		CProbabilityDistribution state( 40, 17, 7, 2 );
		state.DrawCards( 2 );
		table.AddString( "on the play" );
		table.AddProbability( state.TakeMeasurement( 
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands >= 3;
			}
		) );
		state.DrawCard();
		table.AddString( "on the draw" );
		table.AddProbability( state.TakeMeasurement( 
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands >= 3;
			}
		) );

		table.Print();
	}

	{
		CTable table( "Average turn on which a keyrune will be playable, if not on turn three, with a two-land starting hand." );

		const int turn = 3;

		double turn_on_the_play;
		{
			const int hand_size = 7 + turn - 1;

			CProbabilityDistribution state( 40, 17, hand_size, 2 );
			state.DrawUntil( 
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 3;
				}
			);
			turn_on_the_play = state.TakeMeasurement( 
				[ turn, hand_size ]( const CProbabilityDistributionEntry& entry )
				{
					return turn + entry.m_player.m_hand.GetSize() - hand_size;
				}
			);
		}

		double turn_on_the_draw;
		{
			const int hand_size = 7 + turn;

			CProbabilityDistribution state( 40, 17, hand_size, 2 );
			state.DrawUntil( 
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 3;
				}
			);
			turn_on_the_draw = state.TakeMeasurement( 
				[ turn, hand_size ]( const CProbabilityDistributionEntry& entry )
				{
					return turn + entry.m_player.m_hand.GetSize() - hand_size;
				}
			);
		}

		table.AddAddendum( "average turn" );
		table.AddNumber( ( turn_on_the_play + turn_on_the_draw ) / 2.0, 2 );

		table.Print();
	}

	{
		CProbabilityDistribution state( 40, 17 );
		state.DrawCards( 6 );
		state.DrawCards( 2 );
		CTable table( "Chance of having three mana on turn three with a six-card starting hand." );
		table.AddColumn( "" );
		table.AddColumn( "probability" );
		table.AddString( "on the play" );
		table.AddProbability( state.TakeMeasurement( 
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands >= 3;
			}
		) );
		state.DrawCards( 1 );
		table.AddString( "on the draw" );
		table.AddProbability( state.TakeMeasurement( 
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands >= 3;
			}
		) );
		table.Print();
	}

	{
		const int number_of_rats = 1;

		CTable table( "Chance of having Pack Rat in hand on a given turn, playing a deck with nothing but %d Pack Rats and Swamps, if you mulligan until you either get Pack Rat or are down to one card.", number_of_rats );
		table.AddColumn( "turn" );
		table.AddColumn( "on the play" );
		table.AddColumn( "on the draw" );

		CProbabilityDistribution state( 40, 40 - number_of_rats );
		state.DrawCards( 7 );
		state.MulliganIf( 
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_spells == 0 && entry.m_player.m_hand.GetSize() > 1;
			}
		);

		for ( int turn = 1; turn < 6; ++turn )
		{
			table.AddInteger( turn );
			const double chance_on_the_play = state.TakeMeasurement( 
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells > 0;
				}
			);
			table.AddProbability( chance_on_the_play );

			state.DrawCard();

			const double chance_on_the_draw = state.TakeMeasurement( 
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells > 0;
				}
			);
			table.AddProbability( chance_on_the_draw );
		}

		table.Print();
	}

	{
		CTable table( "Chance of having Pack Rat in hand on turn two, playing a deck with nothing but the given number of Pack Rats plus Swamps, if you mulligan until you either get Pack Rat or are down to one card." );
		table.AddColumn( "rats" );
		table.AddColumn( "probability" );

		for ( int number_of_rats = 1; number_of_rats < 4; ++number_of_rats )
		{
			table.AddInteger( number_of_rats );

			CProbabilityDistribution state( 40, 40 - number_of_rats );
			state.DrawCards( 7 );
			state.MulliganIf( 
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells == 0 && entry.m_player.m_hand.GetSize() > 1;
				}
			);
			state.DrawCard();

			const double chance_on_the_play = state.TakeMeasurement( 
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells > 0;
				}
			);

			state.DrawCard();

			const double chance_on_the_draw = state.TakeMeasurement( 
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells > 0;
				}
			);

			table.AddProbability( ( chance_on_the_play + chance_on_the_draw ) / 2 );
		}

		table.Print();
	}

	{
		CTable match_table( "Chance of winning a match, assuming you win every game in which you get Pack Rat by turn two." );
		match_table.AddColumn( "rats" );
		match_table.AddColumn( "probability" );

		CTable tournament_table( "Chance of winning five matches in a row, assuming you win every game in which you get Pack Rat by turn two." );
		tournament_table.AddColumn( "rats" );
		tournament_table.AddColumn( "probability" );

		for ( int number_of_rats = 1; number_of_rats < 4; ++number_of_rats )
		{
			match_table.AddInteger( number_of_rats );
			tournament_table.AddInteger( number_of_rats );

			CProbabilityDistribution state( 40, 40 - number_of_rats );
			state.DrawCards( 7 );
			state.MulliganIf( 
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells == 0 && entry.m_player.m_hand.GetSize() > 1;
				}
			);
			state.DrawCard();

			const double win_chance_on_the_play = state.TakeMeasurement( 
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells > 0;
				}
			);

			state.DrawCard();

			const double win_chance_on_the_draw = state.TakeMeasurement( 
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells > 0;
				}
			);

			const double probability_of_winning_game_one = ( win_chance_on_the_play + win_chance_on_the_draw  ) / 2;

			// possible match outcomes:
			// - WW:	match win
			// - WLW:	match win
			// - WLL:	match loss
			// - LWW:	match win
			// - LWL:	match loss
			// - LL:	match loss
			const double probability_of_WW = probability_of_winning_game_one * win_chance_on_the_draw;
			const double probability_of_WLW = probability_of_winning_game_one * ( 1.0 - win_chance_on_the_draw ) * win_chance_on_the_play;
			const double probability_of_LWW = ( 1.0 - probability_of_winning_game_one ) * win_chance_on_the_play * win_chance_on_the_draw;

			const double match_win_chance = probability_of_WW + probability_of_WLW + probability_of_LWW;

			match_table.AddProbability( match_win_chance );

			const double five_rounds_undefeated_chance = pow( match_win_chance, 5 );
			tournament_table.AddProbability( five_rounds_undefeated_chance );
		}

		match_table.Print();

		tournament_table.Print();
	}
}