#include "magic_probability_toolkit.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

namespace A08_P01
{

typedef CardSet::TwoLandColorsWithFixing Card;
typedef CardSetTypes< Card >::CPile CPile;
typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

void part_01()
{
	CPile deck;
	deck.m_cards[ Card::LandC ] = 8;
	deck.m_cards[ Card::LandD ] = 8;
	deck.m_cards[ Card::Fixing ] = 1;
	deck.m_cards[ Card::Spell ] = 40 - deck.m_cards[ Card::LandC ] - deck.m_cards[ Card::LandD ] - deck.m_cards[ Card::Fixing ];

	auto store_number_of_gates = []( CProbabilityDistributionEntry& entry )
	{
		entry.m_user_data = entry.m_player.m_hand.m_cards[ Card::Fixing ];
	};

	{
		CTable table( "Chance of casting a card of the listed cost on turn 2. Deck contains 8 each of two basic lands plus 1 gate." );
		table.AddColumn( "casting cost" );
		table.AddColumn( "probability" );
		table.AddColumn( "prob of failure relative to baseline" );

		CProbabilityDistribution state_on_the_play( deck );
		state_on_the_play.DrawCards( 7 );
		state_on_the_play.ForEach( store_number_of_gates );
		state_on_the_play.DrawCard();

		CProbabilityDistribution state_on_the_draw( deck );
		state_on_the_draw.DrawCards( 7 );
		state_on_the_draw.DrawCard();
		state_on_the_draw.ForEach( store_number_of_gates );
		state_on_the_draw.DrawCard();

		double baseline_probability;

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_first_turn = std::min( int( entry.m_user_data ), 1 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_first_turn < 1 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_first_turn < 2 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			baseline_probability = chance_of_casting;

			table.AddString( "1C" );
			table.AddProbability( chance_of_casting );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_first_turn = std::min( int( entry.m_user_data ), 1 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_first_turn < 1 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_first_turn < 1 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_first_turn < 2 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			table.AddString( "CD" );
			table.AddProbability( chance_of_casting );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_first_turn = std::min( int( entry.m_user_data ), 1 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_first_turn < 2 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			table.AddString( "CC" );
			table.AddProbability( chance_of_casting );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		table.Print();
	}

	{
		CTable table( "Chance of casting a card of the listed cost on turn 3. Deck contains 8 each of two basic lands plus 1 gate." );
		table.AddColumn( "casting cost" );
		table.AddColumn( "probability" );
		table.AddColumn( "prob of failure relative to baseline" );

		CProbabilityDistribution state_on_the_play( deck );
		state_on_the_play.DrawCards( 7 );
		state_on_the_play.DrawCard();
		state_on_the_play.ForEach( store_number_of_gates );
		state_on_the_play.DrawCard();

		CProbabilityDistribution state_on_the_draw( deck );
		state_on_the_draw.DrawCards( 7 );
		state_on_the_draw.DrawCards( 2 );
		state_on_the_draw.ForEach( store_number_of_gates );
		state_on_the_draw.DrawCard();
		
		double baseline_probability;

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_second_turn = std::min( int( entry.m_user_data ), 2 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_second_turn < 1 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_second_turn < 3 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			baseline_probability = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;
		}

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_second_turn = std::min( int( entry.m_user_data ), 2 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_second_turn < 3 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			table.AddString( "1HH" );
			table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_second_turn = std::min( int( entry.m_user_data ), 2 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_second_turn < 1 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_second_turn < 3 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			table.AddString( "2C" );
			table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_second_turn = std::min( int( entry.m_user_data ), 2 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_second_turn < 1 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_second_turn < 1 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_second_turn < 3 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			table.AddString( "1CD" );
			table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_second_turn = std::min( int( entry.m_user_data ), 2 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_second_turn < 2 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_second_turn < 3 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			table.AddString( "1CC" );
			table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		table.Print();
	}

	{
		CTable table( "Chance of casting a card of the listed cost on turn 4. Deck contains 8 each of two basic lands plus 1 gate." );
		table.AddColumn( "casting cost" );
		table.AddColumn( "probability" );
		table.AddColumn( "prob of failure relative to baseline" );

		CProbabilityDistribution state_on_the_play( deck );
		state_on_the_play.DrawCards( 7 );
		state_on_the_play.DrawCards( 2 );
		state_on_the_play.ForEach( store_number_of_gates );
		state_on_the_play.DrawCard();

		CProbabilityDistribution state_on_the_draw( deck );
		state_on_the_draw.DrawCards( 7 );
		state_on_the_draw.DrawCards( 3 );
		state_on_the_draw.ForEach( store_number_of_gates );
		state_on_the_draw.DrawCard();
		
		double baseline_probability;

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_third_turn = std::min( int( entry.m_user_data ), 3 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_third_turn < 4 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			baseline_probability = chance_of_casting;

			table.AddString( "3C" );
			table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_third_turn = std::min( int( entry.m_user_data ), 3 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_third_turn < 1 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_third_turn < 1 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_third_turn < 4 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			table.AddString( "2CD" );
			table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}
		
		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_third_turn = std::min( int( entry.m_user_data ), 3 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_third_turn < 2 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_third_turn < 4 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			table.AddString( "2CC" );
			table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_third_turn = std::min( int( entry.m_user_data ), 3 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_third_turn < 2 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_third_turn < 2 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_third_turn < 4 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			table.AddString( "CCDD" );
			table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_third_turn = std::min( int( entry.m_user_data ), 3 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_third_turn < 3 ) return false;
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + gates_in_play_on_third_turn < 4 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			table.AddString( "1CCC" );
			table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		{
			auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
			{
				const int gates_in_play_on_third_turn = std::min( int( entry.m_user_data ), 3 );
				if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + gates_in_play_on_third_turn < 4 ) return false;
				return true;
			};
			
			const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
			const double chance_of_casting = ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0;

			table.AddString( "CCCC" );
			table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			table.AddProbability( 1.0 - chance_of_casting / baseline_probability);
		}

		table.Print();
	}

	{
		CTable table( "Chance of casting a card of the listed cost on turn 2. Deck contains 17 lands total, of which the given number are C lands and the remainder are D lands." );
		table.AddColumn( "C lands" );
		table.AddColumn( "D lands" );
		table.AddColumn( "cost HH" );
		table.AddColumn( "cost 1C" );
		table.AddColumn( "cost CD" );
		table.AddColumn( "cost CC" );

		for ( int number_of_C_lands = 0; number_of_C_lands < 18; ++number_of_C_lands )
		{
			CPile deck;
			deck.m_cards[ Card::LandC ] = number_of_C_lands;
			deck.m_cards[ Card::LandD ] = 17 - number_of_C_lands;
			deck.m_cards[ Card::Fixing ] = 0;
			deck.m_cards[ Card::Spell ] = 40 - 17;

			table.AddInteger( deck.m_cards[ Card::LandC ] );
			table.AddInteger( deck.m_cards[ Card::LandD ] );

			CProbabilityDistribution state_on_the_play( deck );
			state_on_the_play.DrawCards( 7 );
			state_on_the_play.DrawCard();

			CProbabilityDistribution state_on_the_draw( deck );
			state_on_the_draw.DrawCards( 7 );
			state_on_the_draw.DrawCards( 2 );

			{ // HH
				auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
				{
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] < 2 ) return false;
					return true;
				};
			
				const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
				const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
				table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			}

			{ // 1C
				auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
				{
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] < 1 ) return false;
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] < 2 ) return false;
					return true;
				};
			
				const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
				const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
				table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			}

			{ // CD
				auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
				{
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] < 1 ) return false;
					if ( entry.m_player.m_hand.m_cards[ Card::LandD ] < 1 ) return false;
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] < 2 ) return false;
					return true;
				};
			
				const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
				const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
				table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			}

			{ // CC
				auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
				{
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] < 2 ) return false;
					return true;
				};
			
				const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
				const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
				table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			}
		}

		table.Print();
	}

	{
		CTable table( "Chance of casting a card of the listed cost on turn 3. Deck contains 17 lands total, of which the given number are C lands and the remainder are D lands." );
		table.AddColumn( "C lands" );
		table.AddColumn( "D lands" );
		table.AddColumn( "cost 1HH" );
		table.AddColumn( "cost 2C" );
		table.AddColumn( "cost 1CD" );
		table.AddColumn( "cost 1CC" );

		for ( int number_of_C_lands = 0; number_of_C_lands < 18; ++number_of_C_lands )
		{
			CPile deck;
			deck.m_cards[ Card::LandC ] = number_of_C_lands;
			deck.m_cards[ Card::LandD ] = 17 - number_of_C_lands;
			deck.m_cards[ Card::Fixing ] = 0;
			deck.m_cards[ Card::Spell ] = 40 - 17;

			table.AddInteger( deck.m_cards[ Card::LandC ] );
			table.AddInteger( deck.m_cards[ Card::LandD ] );

			CProbabilityDistribution state_on_the_play( deck );
			state_on_the_play.DrawCards( 7 );
			state_on_the_play.DrawCards( 2 );

			CProbabilityDistribution state_on_the_draw( deck );
			state_on_the_draw.DrawCards( 7 );
			state_on_the_draw.DrawCards( 3 );

			{ // 1HH
				auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
				{
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] < 3 ) return false;
					return true;
				};
			
				const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
				const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
				table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			}

			{ // 2C
				auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
				{
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] < 1 ) return false;
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] < 3 ) return false;
					return true;
				};
			
				const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
				const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
				table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			}

			{ // 1CD
				auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
				{
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] < 1 ) return false;
					if ( entry.m_player.m_hand.m_cards[ Card::LandD ] < 1 ) return false;
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] < 3 ) return false;
					return true;
				};
			
				const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
				const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
				table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			}

			{ // 1CC
				auto can_cast_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
				{
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] < 2 ) return false;
					if ( entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] < 3 ) return false;
					return true;
				};
			
				const double chance_of_casting_on_the_play = state_on_the_play.TakeMeasurement( can_cast_measurement );
				const double chance_of_casting_on_the_draw = state_on_the_draw.TakeMeasurement( can_cast_measurement );
				table.AddProbability( ( chance_of_casting_on_the_play + chance_of_casting_on_the_draw ) / 2.0 );
			}
		}

		table.Print();
	}
}

};

namespace A08_P02
{

void part_02()
{
	{
		CTable table( "Average number of additional spells drawn before the end of the game due to thinning land out of deck with various spells. Assume that each spell is in a starting hand with four lands and is cast on the first possible turn." );
		table.AddColumn( "spell" );
		table.AddColumn( "casting turn" );
		table.AddColumn( "lands removed" );
		table.AddColumn( "additional spells" );

		const int casting_turn_table[] = { 2, 4, 2, 1 };
		const int lands_removed_from_deck_table[] = { 1, 3, -1, 1 };
		const char* spell_name_table[] = { "Gatecreeper Vine", "Seek the Horizon", "Mana Severance", "Arid Mesa" };
		const int end_of_game_turn = 10;

		for ( int table_idx = 0; table_idx < std::extent< decltype( casting_turn_table ) >::value; ++table_idx )
		{
			const int casting_turn = casting_turn_table[ table_idx ];
			const int lands_removed_from_deck = lands_removed_from_deck_table[ table_idx ];

			double spells_at_game_end_without_thinning;
			{
				CProbabilityDistribution state( 40, 17, 7, 4 );
				state.DrawCards( casting_turn );
				state.DrawCards( end_of_game_turn - casting_turn );
				spells_at_game_end_without_thinning = state.TakeMeasurement( 
					[]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.m_spells;
					}
				);
			}

			double spells_at_game_end_with_thinning;
			{
				CProbabilityDistribution state( 40, 17, 7, 4 );
				state.DrawCards( casting_turn );
				state.ForEach( 
					[ lands_removed_from_deck ]( CProbabilityDistributionEntry& entry )
					{
						if ( lands_removed_from_deck > 0 )
						{
							assert( entry.m_player.m_library.m_lands >= lands_removed_from_deck );
							entry.m_player.m_hand.m_lands += lands_removed_from_deck;
							entry.m_player.m_library.m_lands -= lands_removed_from_deck;
						}
						else
						{
							entry.m_player.m_hand.m_lands += entry.m_player.m_library.m_lands;
							entry.m_player.m_library.m_lands = 0;
						}
					}
				);
				state.DrawCards( end_of_game_turn - casting_turn );
				spells_at_game_end_with_thinning = state.TakeMeasurement( 
					[]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.m_spells;
					}
				);
			}

			table.AddString( spell_name_table[ table_idx ] );
			table.AddInteger( casting_turn_table[ table_idx ] );
			if ( lands_removed_from_deck_table[ table_idx ] < 0 )
			{
				table.AddString( "all" );
			}
			else
			{
				table.AddInteger( lands_removed_from_deck_table[ table_idx ] );
			}
			table.AddNumber( spells_at_game_end_with_thinning - spells_at_game_end_without_thinning, 2 );
		}

		table.Print();
	}
}

};

void article_08()
{
	A08_P01::part_01();
	A08_P02::part_02();
}