#include "magic_probability_toolkit.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

void article_09()
{
	{
		CProbabilityDistribution state( 40, 17, 7, 3 );
	
		CTable table( "Chance of having missed a land drop by a given turn." );
		table.AddColumn( "turn" );
		table.AddColumn( "probability" );
		for ( int turn = 1; turn < 13; ++turn )
		{
			const double chance_of_hitting_land_drop_on_the_play = state.TakeMeasurement(
				[ turn ]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= turn;
				}
			);
			state.DrawCard();
			const double chance_of_hitting_land_drop_on_the_draw = state.TakeMeasurement(
				[ turn ]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= turn;
				}
			);
			table.AddInteger( turn );
			table.AddProbability( 1.0 - ( chance_of_hitting_land_drop_on_the_play + chance_of_hitting_land_drop_on_the_draw ) / 2.0 );
		}
		table.Print();
	}

	{
		CProbabilityDistribution state( 40, 17, 7, 3 );

		CTable table_lands_in_play( "Average number of lands in play on a given turn." );
		table_lands_in_play.AddColumn( "turn" );
		table_lands_in_play.AddColumn( "lands in play" );

		CTable table_growth_rate( "Rate of mana growth." );
		table_growth_rate.AddColumn( "turn" );
		table_growth_rate.AddColumn( "lands per turn" );

		double previous_lands_in_play = 0.0;
		for ( int turn = 1; turn < 15; ++turn )
		{
			const double lands_in_play_on_the_play = state.TakeMeasurement(
				[ turn ]( const CProbabilityDistributionEntry& entry )
				{
					return std::min( turn, entry.m_player.m_hand.m_lands );
				}
			);
			state.DrawCard();
			const double lands_in_play_on_the_draw = state.TakeMeasurement(
				[ turn ]( const CProbabilityDistributionEntry& entry )
				{
					return std::min( turn, entry.m_player.m_hand.m_lands );
				}
			);
			table_lands_in_play.AddInteger( turn );

			const double lands_in_play = ( lands_in_play_on_the_play + lands_in_play_on_the_draw ) / 2.0;
			table_lands_in_play.AddNumber( lands_in_play );

			table_growth_rate.AddInteger( turn );
			table_growth_rate.AddNumber( lands_in_play - previous_lands_in_play, 2 );

			previous_lands_in_play = lands_in_play;
		}

		table_lands_in_play.Print();
		table_growth_rate.Print();
	}

	{
		CTable table( "Average casting turn of cards with various converted mana costs." );
		table.AddColumn( "mana cost" );
		table.AddColumn( "casting turn" );

		for ( int casting_cost = 1; casting_cost < 12; ++casting_cost )
		{
			double turn_cost_on_the_play;
			{
				CProbabilityDistribution state( 40, 17, 7, 3 );
				state.DrawUntil(
					[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
					{
						const int turn = entry.m_player.m_hand.GetSize() - 7 + 1;
						return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
					}
				);
				turn_cost_on_the_play = state.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry ) -> int
					{
						const int turn = entry.m_player.m_hand.GetSize() - 7 + 1;
						return turn;
					}
				);
			}

			double turn_cost_on_the_draw;
			{
				CProbabilityDistribution state( 40, 17, 7, 3 );
				state.DrawUntil(
					[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
					{
						const int turn = entry.m_player.m_hand.GetSize() - 7;
						return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
					}
				);
				turn_cost_on_the_draw = state.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry ) -> int
					{
						const int turn = entry.m_player.m_hand.GetSize() - 7;
						return turn;
					}
				);
			}

			table.AddInteger( casting_cost );
			table.AddNumber( ( turn_cost_on_the_play + turn_cost_on_the_draw ) / 2.0 );
		}

		table.Print();
	}

	for ( int casting_cost = 1; casting_cost < 12; ++casting_cost )
	{
		CTable table( "Casting turn distribution of a spell with converted mana cost of %d.", casting_cost );
		table.AddColumn( "turn" );
		table.AddColumn( "probability" );

		{
			double casting_turn_on_the_play_table[ 100 ];
			{
				CProbabilityDistribution state( 40, 17, 7, 3 );
				state.DrawUntil(
					[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
					{
						const int turn = entry.m_player.m_hand.GetSize() - 7 + 1;
						return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
					}
				);
				state.BinDistribution(
					casting_turn_on_the_play_table,
					std::extent< decltype( casting_turn_on_the_play_table ) >::value,
					[]( const CProbabilityDistributionEntry& entry ) -> int
					{
						const int turn = entry.m_player.m_hand.GetSize() - 7 + 1;
						return turn;
					}
				);
			}

			double casting_turn_on_the_draw_table[ 100 ];
			{
				CProbabilityDistribution state( 40, 17, 7, 3 );
				state.DrawUntil(
					[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
					{
						const int turn = entry.m_player.m_hand.GetSize() - 7;
						return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
					}
				);
				state.BinDistribution(
					casting_turn_on_the_draw_table,
					std::extent< decltype( casting_turn_on_the_draw_table ) >::value,
					[]( const CProbabilityDistributionEntry& entry ) -> int
					{
						const int turn = entry.m_player.m_hand.GetSize() - 7;
						return turn;
					}
				);
			}

			for ( int turn = 0; turn < std::extent< decltype( casting_turn_on_the_draw_table ) >::value; ++turn )
			{
				const double probability = ( casting_turn_on_the_play_table[ turn ] + casting_turn_on_the_draw_table[ turn ] ) / 2.0;
				if ( probability > 0.0 )
				{
					table.AddInteger( turn );
					table.AddProbability( probability );
				}
			}

			table.Print();
		}
	}
}