#include "magic_probability_toolkit.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

void article_10()
{
	const int prefered_lower_bound = 3;
	const int prefered_upper_bound = 5;

	int hand_sizes[] = { 4, 5, 6, 7 };

	double land_distribution_table[ std::extent< decltype( hand_sizes ) >::value ][ 8 ];
	double average_land_count_table[ std::extent< decltype( hand_sizes ) >::value ];
	double land_situation_table[ std::extent< decltype( hand_sizes ) >::value ][ 3 ];

	for ( int idx = 0; idx < std::extent< decltype( hand_sizes ) >::value; ++idx )
	{
		CProbabilityDistribution state( 40, 17 );
		state.DrawCards( hand_sizes[ idx  ] );

		state.BinDistribution(
			land_distribution_table[ idx ],
			std::extent< decltype( land_distribution_table ), 1 >::value,
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands;
			}
		);

		average_land_count_table[ idx ] = state.TakeMeasurement(
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands;
			}
		);

		state.BinDistribution(
			land_situation_table[ idx ],
			std::extent< decltype( land_situation_table ), 1 >::value,
			[ prefered_lower_bound, prefered_upper_bound ]( const CProbabilityDistributionEntry& entry ) -> int
			{
				if ( entry.m_player.m_hand.m_lands < prefered_lower_bound ) return 0;
				if ( entry.m_player.m_hand.m_lands > prefered_upper_bound ) return 2;
				return 1;
			}
		);
	}

	{
		CTable table( "Land count distribution for various starting hand sizes." );
		table.AddColumn( "land" );
		table.AddColumn( "4-cards" );
		table.AddColumn( "5-cards" );
		table.AddColumn( "6-cards" );
		table.AddColumn( "7-cards" );
		for ( int lands = 0; lands < std::extent< decltype( land_distribution_table ), 1 >::value; ++lands )
		{
			table.AddInteger( lands );
			table.AddProbability( land_distribution_table[ 0 ][ lands ], 1, true );
			table.AddProbability( land_distribution_table[ 1 ][ lands ], 1, true );
			table.AddProbability( land_distribution_table[ 2 ][ lands ], 1, true );
			table.AddProbability( land_distribution_table[ 3 ][ lands ], 1, true );
		}
		table.Print();
	}

	{
		CTable table( "Average land count for various starting hand sizes." );
		table.AddColumn( "hand" );
		table.AddColumn( "average land" );
		for ( int idx = std::extent< decltype( hand_sizes ) >::value; idx--; )
		{
			table.AddInteger( hand_sizes[ idx  ] );
			table.AddNumber( average_land_count_table[ idx  ], 3 );
		}
		table.Print();
	}

	{
		CTable table( "Land situation distribution for various starting hand sizes." );
		table.AddColumn( "hand" );
		table.AddColumn( "preferred" );
		table.AddColumn( "screwed or flooded" );
		for ( int idx = std::extent< decltype( hand_sizes ) >::value; idx--; )
		{
			table.AddInteger( hand_sizes[ idx  ] );
			table.AddProbability( land_situation_table[ idx ][ 1 ] );
			table.AddProbability( land_situation_table[ idx ][ 0 ] + land_situation_table[ idx ][ 2 ] );
		}
		table.Print();
	}

	auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry )
	{
		return entry.m_player.m_hand.m_lands >= 4 && entry.m_player.m_hand.m_spells >= 4;
	};

	for ( int starting_hand_size = 7; starting_hand_size >= 6; --starting_hand_size )
	{
		CTable table( "Chance of a competitive start with various land counts in a %d-card hand.", starting_hand_size );
		table.AddColumn( "land" );
		table.AddColumn( "spells" );
		table.AddColumn( "probability" );

		for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
		{
			CProbabilityDistribution state( 40, 17, starting_hand_size, lands_in_starting_hand );
			state.DrawCards( 4 );

			const double chance_of_competitive_start = state.TakeMeasurement( competitive_start_measurement );
			table.AddInteger( lands_in_starting_hand );
			table.AddInteger( starting_hand_size - lands_in_starting_hand );
			table.AddProbability( chance_of_competitive_start );
		}
		
		table.Print();
	}

	double overall_chance_of_competitive_start_with_six_card_hand;

	{
		CTable table( "Overall chance of a competitive start a 6-card hand." );
		table.AddColumn( "land" );
		table.AddColumn( "prob of hand" );
		table.AddColumn( "" );
		table.AddColumn( "prob of comp start" );
		table.AddColumn( "" );
		table.AddColumn( "" );

		double overall_chance_of_competitive_start = 0.0;

		for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= 6; ++lands_in_starting_hand )
		{
			table.AddInteger( lands_in_starting_hand );

			CProbabilityDistribution state_hand( 40, 17 );
			state_hand.DrawCards( 6 );
			const double chance_of_hand = state_hand.TakeMeasurement(
				[ lands_in_starting_hand ]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands == lands_in_starting_hand;
				}
			);
			table.AddProbability( chance_of_hand );

			table.AddString( "x" );

			CProbabilityDistribution state_competitive_start( 40, 17, 6, lands_in_starting_hand );
			state_competitive_start.DrawCards( 4 );
			const double chance_of_competitive_start = state_competitive_start.TakeMeasurement( competitive_start_measurement );
			table.AddProbability( chance_of_competitive_start );

			table.AddString( "=" );

			table.AddProbability( chance_of_hand * chance_of_competitive_start );

			overall_chance_of_competitive_start += chance_of_hand * chance_of_competitive_start;
		}

		table.AddAddendum( "total probability of competitive start" );
		table.AddProbability( overall_chance_of_competitive_start );
		
		table.Print();

		overall_chance_of_competitive_start_with_six_card_hand = overall_chance_of_competitive_start;
	}

	{
		const int starting_hand_size = 7;

		CTable table( "Comparison of competitive start chance between various land counts in a 7-card hand and the average 6-card hand." );
		table.AddColumn( "land" );
		table.AddColumn( "prob of comp start" );
		table.AddColumn( "" );
		table.AddColumn( "6-card hand" );
		table.AddColumn( "result" );

		for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
		{
			CProbabilityDistribution state( 40, 17, starting_hand_size, lands_in_starting_hand );
			state.DrawCards( 4 );

			const double chance_of_competitive_start = state.TakeMeasurement( competitive_start_measurement );
			const bool keep = chance_of_competitive_start >= overall_chance_of_competitive_start_with_six_card_hand;

			table.AddInteger( lands_in_starting_hand );
			table.AddProbability( chance_of_competitive_start );
			table.AddString( keep ? ">" : "<" );
			table.AddProbability( overall_chance_of_competitive_start_with_six_card_hand );
			table.AddString( keep ? "keep" : "mulligan" );
		}
		
		table.Print();
	}
}