#include "magic_probability_toolkit.h"
#include "mulligan_rules.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;
using namespace mlr;

void article_11()
{
	if ( false )
	{
		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry )
		{
			return entry.m_player.m_hand.m_lands >= 4 && entry.m_player.m_hand.m_spells >= 4;
		};

		CCalculateMulliganRulesResults results = CalculateMulliganRulesIgnoringColor( Draw, 4, competitive_start_measurement );
		results.PrintEntireCalculation( "Mulligan rules calculation on the draw with standard competitive start defintion." );
		results.PrintTable( "Mulligan rules on the draw with standard competitive start definition" );
		results.PrintComplexTable( "Mulligan rules on the draw with standard competitive start definition" );
	}

	auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry )
	{
		return entry.m_player.m_hand.m_lands >= 4 && entry.m_player.m_hand.m_spells >= 4;
	};

	double overall_chance_of_competitive_start_table[ 8 ];
	bool mulligan_rules_table[ 8 ][ 8 ];
	enum { Mulligan, ConsiderMulligan, Keep } complex_mulligan_rules_table[ 8 ][ 8 ];
	double probability_of_mulligan[ 8 ];

	overall_chance_of_competitive_start_table[ 2 ] = 0.0;

	for ( int starting_hand_size = 3; starting_hand_size < 8; ++starting_hand_size )
	{
		double m_chance_of_competitive_start_table[ 8 ];
		double m_chance_of_drawing_hand_table[ 8 ];

		probability_of_mulligan[ starting_hand_size ] = 0.0;

		{
			CTable table( "Chance of a competitive start with various land counts in a %d-card hand.", starting_hand_size );
			table.AddColumn( "land" );
			table.AddColumn( "spells" );
			table.AddColumn( "probability" );

			for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
			{
				CProbabilityDistribution state( 40, 17, starting_hand_size, lands_in_starting_hand );
				state.DrawCards( 4 );

				const double chance_of_competitive_start = state.TakeMeasurement( competitive_start_measurement );
				table.AddInteger( lands_in_starting_hand );
				table.AddInteger( starting_hand_size - lands_in_starting_hand );
				table.AddProbability( chance_of_competitive_start );

				m_chance_of_competitive_start_table[ lands_in_starting_hand ] = chance_of_competitive_start;
			}

			{
				CProbabilityDistribution state( 40, 17 );
				state.DrawCards( starting_hand_size );
				state.BinDistribution(
					m_chance_of_drawing_hand_table,
					8,
					[]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.m_lands;
					}
				);
			}
		
			table.Print();
		}

		{
			double mulligan_adjusted_chance_of_competitive_start = 0.0;
			double m_unadjusted_chance_of_competitive_start = 0.0;

			CTable comparison_table( "Comparison of competitive start chance between various land counts in a %d-card hand and the average %d-card hand.", starting_hand_size, starting_hand_size - 1 );
			comparison_table.AddColumn( "land" );
			comparison_table.AddColumn( "prob of comp start" );
			comparison_table.AddColumn( "" );
			comparison_table.AddColumn( "%d-card hand", starting_hand_size - 1 );
			comparison_table.AddColumn( "result" );

			CTable overall_chance_table( "Overall chance of a competitive start a %d-card hand with adjusted probability on mulliganed hands.", starting_hand_size );
			overall_chance_table.AddColumn( "land" );
			overall_chance_table.AddColumn( "prob of hand" );
			overall_chance_table.AddColumn( "" );
			overall_chance_table.AddColumn( "prob of comp start" );
			overall_chance_table.AddColumn( "" );
			overall_chance_table.AddColumn( "" );

			for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
			{
				mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] = m_chance_of_competitive_start_table[ lands_in_starting_hand ] < overall_chance_of_competitive_start_table[ starting_hand_size - 1 ];
				if ( abs( m_chance_of_competitive_start_table[ lands_in_starting_hand ] - overall_chance_of_competitive_start_table[ starting_hand_size - 1 ] ) < 0.1f )
				{
					complex_mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] = ConsiderMulligan;
				}
				else if ( mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] )
				{
					complex_mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] = Mulligan;
				}
				else
				{
					complex_mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] = Keep;
				}

				if ( mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] )
				{
					probability_of_mulligan[ starting_hand_size ] += m_chance_of_drawing_hand_table[ lands_in_starting_hand ];
				}

				double chance = mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ]
					? overall_chance_of_competitive_start_table[ starting_hand_size - 1 ]
					: m_chance_of_competitive_start_table[ lands_in_starting_hand ];

				const double result = m_chance_of_drawing_hand_table[ lands_in_starting_hand ] * chance;
				mulligan_adjusted_chance_of_competitive_start += result;

				m_unadjusted_chance_of_competitive_start += m_chance_of_drawing_hand_table[ lands_in_starting_hand ] * m_chance_of_competitive_start_table[ lands_in_starting_hand ];

				comparison_table.AddInteger( lands_in_starting_hand );
				comparison_table.AddProbability( m_chance_of_competitive_start_table[ lands_in_starting_hand ] );
				comparison_table.AddString( mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] ? "<" : ">" );
				comparison_table.AddProbability( overall_chance_of_competitive_start_table[ starting_hand_size - 1 ] );
				comparison_table.AddString( mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] ? "mulligan" : "keep" );

				overall_chance_table.AddInteger( lands_in_starting_hand );
				overall_chance_table.AddProbability( m_chance_of_drawing_hand_table[ lands_in_starting_hand ] );
				overall_chance_table.AddString( "x" );
				overall_chance_table.AddProbability( chance );
				overall_chance_table.AddString( "=" );
				overall_chance_table.AddProbability( result );
			}

			overall_chance_table.AddAddendum( "mulligan-adjusted chance of a competitive start" );
			overall_chance_table.AddProbability( mulligan_adjusted_chance_of_competitive_start );
			overall_chance_table.AddAddendum( "unadjusted chance of a competitive start" );
			overall_chance_table.AddProbability( m_unadjusted_chance_of_competitive_start );

			comparison_table.Print();

			overall_chance_table.Print();

			overall_chance_of_competitive_start_table[ starting_hand_size ] = mulligan_adjusted_chance_of_competitive_start;
		}
	}

	{
		CTable table( "Chance of keeping with at a given hand size." );
		table.AddColumn( "hand size" );
		table.AddColumn( "prob of mulligan" );

		for ( int starting_hand_size = 7; starting_hand_size > 3; --starting_hand_size )
		{
			table.AddInteger( starting_hand_size );
			table.AddProbability( 1.0 - probability_of_mulligan[ starting_hand_size ] );
		}

		table.Print();
	}
	
	{
		CTable table( "Distribution of land count using mulligans and without using mulligans." );
		table.AddColumn( "lands" );
		table.AddColumn( "without using mulligans" );
		table.AddColumn( "using mulligans" );

		double without_mulligan_hand_size_table[ 8 ];
		double with_mulligan_hand_size_table[ 8 ];

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::NeverMulligan );
			state.BinDistribution(
				without_mulligan_hand_size_table,
				std::extent< decltype( without_mulligan_hand_size_table ) >::value,
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands;
				}
			);
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			state.BinDistribution(
				with_mulligan_hand_size_table,
				std::extent< decltype( with_mulligan_hand_size_table ) >::value,
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands;
				}
			);
		}

		for ( int hand_size = 0; hand_size < std::extent< decltype( without_mulligan_hand_size_table ) >::value; ++hand_size )
		{
			table.AddInteger( hand_size );
			table.AddProbability( without_mulligan_hand_size_table[ hand_size ] );
			table.AddProbability( with_mulligan_hand_size_table[ hand_size ] );
		}

		table.Print();
	}

	{
		CTable table( "Average land count in a starting hand using mulligans and without using mulligans." );
		table.AddColumn( "hand" );
		table.AddColumn( "lands" );

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::NeverMulligan );
			const double lands = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands;
				}
			);
			table.AddString( "without mulligans" );
			table.AddNumber( lands, 2 );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			const double lands = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands;
				}
			);
			table.AddString( "with mulligans" );
			table.AddNumber( lands, 2 );
		}

		table.Print();
	}
	
	{
		CTable table( "Mana situation using mulligans and without using mulligans." );
		table.AddColumn( "mana situation" );
		table.AddColumn( "without using mulligans" );
		table.AddColumn( "using mulligans" );

		double without_mulligan_mana_situation_table[ 3 ];
		double with_mulligan_mana_situation_table[ 3 ];

		auto mana_situation_binner = []( const CProbabilityDistributionEntry& entry ) -> int
		{
			if ( entry.m_player.m_hand.m_lands < 3 ) return 0;
			if ( entry.m_player.m_hand.m_lands < 6 ) return 1;
			return 2;
		};

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::NeverMulligan );
			state.BinDistribution(
				without_mulligan_mana_situation_table,
				std::extent< decltype( without_mulligan_mana_situation_table ) >::value,
				mana_situation_binner
			);
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			state.BinDistribution(
				with_mulligan_mana_situation_table,
				std::extent< decltype( with_mulligan_mana_situation_table ) >::value,
				mana_situation_binner
			);
		}

		table.AddString( "screwed" );
		table.AddProbability( without_mulligan_mana_situation_table[ 0 ] );
		table.AddProbability( with_mulligan_mana_situation_table[ 0 ] );
		table.AddString( "preferred" );
		table.AddProbability( without_mulligan_mana_situation_table[ 1 ] );
		table.AddProbability( with_mulligan_mana_situation_table[ 1 ] );
		table.AddString( "flooded" );
		table.AddProbability( without_mulligan_mana_situation_table[ 2 ] );
		table.AddProbability( with_mulligan_mana_situation_table[ 2 ] );

		table.Print();
	}

	{
		CTable table( "Average casting turn of cards with various converted mana costs and from various starting-hand situations." );
		table.AddColumn( "mana cost" );
		table.AddColumn( "three-land hand" );
		table.AddColumn( "without mulligans" );
		table.AddColumn( "with mulligans" );

		const int casting_cost_table[] = { 3, 6, 9 };
		for ( int casting_cost_index = 0; casting_cost_index < std::extent< decltype( casting_cost_table ) >::value; ++casting_cost_index )
		{
			const int casting_cost = casting_cost_table[ casting_cost_index ];
			table.AddInteger( casting_cost );

			{
				CProbabilityDistribution state( 40, 17 );
				state.DrawStartingHand( MulliganRules::NeverMulligan );
				state.ForEach(
					[]( CProbabilityDistributionEntry& entry )
					{
						entry.m_user_data = entry.m_player.m_hand.GetSize();
					}
				);
				state.DrawUntil(
					[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
						return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
					}
				);
				const double turn_cost = state.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry ) -> int
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
						return turn;
					}
				);

				table.AddNumber( turn_cost );
			}

			{
				CProbabilityDistribution state( 40, 17, 7, 3 );
				state.ForEach(
					[]( CProbabilityDistributionEntry& entry )
					{
						entry.m_user_data = entry.m_player.m_hand.GetSize();
					}
				);
				state.DrawUntil(
					[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
						return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
					}
				);
				const double turn_cost = state.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry ) -> int
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
						return turn;
					}
				);

				table.AddNumber( turn_cost );
			}

			{
				CProbabilityDistribution state( 40, 17 );
				state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
				state.ForEach(
					[]( CProbabilityDistributionEntry& entry )
					{
						entry.m_user_data = entry.m_player.m_hand.GetSize();
					}
				);
				state.DrawUntil(
					[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
						return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
					}
				);
				const double turn_cost = state.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry ) -> int
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
						return turn;
					}
				);

				table.AddNumber( turn_cost );
			}
		}

		table.Print();
	}
	
	{
		CTable table( "Average cards in a starting hand using mulligans and without using mulligans." );
		table.AddColumn( "hand" );
		table.AddColumn( "cards" );

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::NeverMulligan );
			const double cards = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.GetSize();
				}
			);
			table.AddString( "without mulligans" );
			table.AddNumber( cards, 2 );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			const double cards = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.GetSize();
				}
			);
			table.AddString( "with mulligans" );
			table.AddNumber( cards, 2 );
		}

		table.Print();
	}
	
	{
		CTable table( "Distribution of hand size after mulligans." );
		table.AddColumn( "hand size" );
		table.AddColumn( "probability" );

		double hand_size_table[ 8 ];

		CProbabilityDistribution state( 40, 17 );
		state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
		state.BinDistribution(
			hand_size_table,
			std::extent< decltype( hand_size_table ) >::value,
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.GetSize();
			}
		);

		for ( int hand_size = std::extent< decltype( hand_size_table ) >::value; --hand_size; )
		{
			if ( hand_size_table[ hand_size ] > 0.0 )
			{
				table.AddInteger( hand_size );
				table.AddProbability( hand_size_table[ hand_size ] );
			}
		}

		table.Print();
	}

	{
		CTable table( "Chance of a competitive start will tweaked mulligan m_rules." );
		table.AddColumn( "m_rules" );
		table.AddColumn( "prob of comp start" );

		auto mulligan_rules_tweak_keep_two = []( const mpt::CProbabilityDistributionEntry& entry ) -> bool
		{
			if ( entry.m_player.m_hand.GetSize() == 7 )
			{
				return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 5;
			}
			if ( entry.m_player.m_hand.GetSize() == 6 )
			{
				return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 5;
			}
			if ( entry.m_player.m_hand.GetSize() == 5 )
			{
				return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 4;
			}
			return false;
		};

		auto mulligan_rules_tweak_mull_five = []( const mpt::CProbabilityDistributionEntry& entry ) -> bool
		{
			if ( entry.m_player.m_hand.GetSize() == 7 )
			{
				return entry.m_player.m_hand.m_lands < 3 || entry.m_player.m_hand.m_lands > 4;
			}
			if ( entry.m_player.m_hand.GetSize() == 6 )
			{
				return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 5;
			}
			if ( entry.m_player.m_hand.GetSize() == 5 )
			{
				return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 4;
			}
			return false;
		};

		auto mulligan_rules_tweak_keep_two_mull_five = []( const mpt::CProbabilityDistributionEntry& entry ) -> bool
		{
			if ( entry.m_player.m_hand.GetSize() == 7 )
			{
				return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 4;
			}
			if ( entry.m_player.m_hand.GetSize() == 6 )
			{
				return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 5;
			}
			if ( entry.m_player.m_hand.GetSize() == 5 )
			{
				return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 4;
			}
			return false;
		};

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			state.DrawCards( 4 );
			const double probability = state.TakeMeasurement( competitive_start_measurement );
			table.AddString( "use mulligan m_rules as stated" );
			table.AddProbability( probability );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( mulligan_rules_tweak_mull_five );
			state.DrawCards( 4 );
			const double probability = state.TakeMeasurement( competitive_start_measurement );
			table.AddString( "mulligan five-land hands" );
			table.AddProbability( probability );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( mulligan_rules_tweak_keep_two );
			state.DrawCards( 4 );
			const double probability = state.TakeMeasurement( competitive_start_measurement );
			table.AddString( "keep two-land hands" );
			table.AddProbability( probability );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( mulligan_rules_tweak_keep_two_mull_five );
			state.DrawCards( 4 );
			const double probability = state.TakeMeasurement( competitive_start_measurement );
			table.AddString( "keep two-land hands and mulligan five-land hands" );
			table.AddProbability( probability );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::NeverMulligan );
			state.DrawCards( 4 );
			const double probability = state.TakeMeasurement( competitive_start_measurement );
			table.AddString( "no mulligans" );
			table.AddProbability( probability );
		}

		table.Print();
	}

	{
		CTable table( "Mulligan rules on the draw." );
		table.AddColumn( "   " );
		table.AddColumn( " 0" );
		table.AddColumn( " 1" );
		table.AddColumn( " 2" );
		table.AddColumn( " 3" );
		table.AddColumn( " 4" );
		table.AddColumn( " 5" );
		table.AddColumn( " 6" );
		table.AddColumn( " 7" );
		for ( int starting_hand_size = 7; starting_hand_size > 3; --starting_hand_size )
		{
			table.AddInteger( starting_hand_size );
			for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= 7; ++lands_in_starting_hand )
			{
				if ( lands_in_starting_hand > starting_hand_size )
				{
					table.AddString( "" );
				}
				else
				{
					table.AddString( mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] ? "X" : "-" );
				}
			}
		}
		table.Print();
	}

	{
		CTable table( "Complex mulligan m_rules on the draw." );
		table.AddColumn( "   " );
		table.AddColumn( " 0" );
		table.AddColumn( " 1" );
		table.AddColumn( " 2" );
		table.AddColumn( " 3" );
		table.AddColumn( " 4" );
		table.AddColumn( " 5" );
		table.AddColumn( " 6" );
		table.AddColumn( " 7" );
		for ( int starting_hand_size = 7; starting_hand_size > 3; --starting_hand_size )
		{
			table.AddInteger( starting_hand_size );
			for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= 7; ++lands_in_starting_hand )
			{
				if ( lands_in_starting_hand > starting_hand_size )
				{
					table.AddString( "" );
				}
				else
				{
					static const char* const glyphs[] = { "X", "?", "-" };
					table.AddString( glyphs[ complex_mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] ] );
				}
			}
		}
		table.Print();
	}
}