#include "magic_probability_toolkit.h"
#include "mulligan_rules.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;
using namespace mlr;

void article_12()
{
	auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry )
	{
		return entry.m_player.m_hand.m_lands >= 4 && entry.m_player.m_hand.m_spells >= 4;
	};

	{
		CTable table( "Chance of not hitting 3rd land drop on turn 3 on the play and on the draw." );
		table.AddColumn( "player" );
		table.AddColumn( "probability" );

		const int land_drop = 3;
		const int turn = 3;
		{
			CProbabilityDistribution state( 40, 17, 7, 2 );
			state.DrawCards( turn - 1 );
			{
				const double probability =  state.TakeMeasurement(
					[ land_drop ]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.m_lands >= land_drop;
					}
				);
				table.AddString( "on the play" );
				table.AddProbability( 1.0 - probability );
			}

			state.DrawCard();
			{
				const double probability =  state.TakeMeasurement(
					[ land_drop ]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.m_lands >= land_drop;
					}
				);
				table.AddString( "on the draw" );
				table.AddProbability( 1.0 - probability );
			}
		}
		table.Print();
	}

	{
		CTable table( "Chance of not hitting 4rd land drop by turn 5 on the play and on the draw." );
		table.AddColumn( "player" );
		table.AddColumn( "probability" );

		const int land_drop = 4;
		const int turn = 5;
		{
			CProbabilityDistribution state( 40, 17, 7, 2 );
			state.DrawCards( turn - 1 );
			{
				const double probability = state.TakeMeasurement(
					[ land_drop ]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.m_lands >= land_drop;
					}
				);
				table.AddString( "on the play" );
				table.AddProbability( 1.0 - probability );
			}

			state.DrawCard();
			{
				const double probability = state.TakeMeasurement(
					[ land_drop ]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.m_lands >= land_drop;
					}
				);
				table.AddString( "on the draw" );
				table.AddProbability( 1.0 - probability );
			}
		}
		table.Print();
	}

	{
		CTable table( "Chance of a competitive start for various 7-card hands, comparing on the play and on the draw." );
		table.AddColumn( "lands" );
		table.AddColumn( "on the play" );
		table.AddColumn( "on the draw" );

		const int starting_hand_size = 7;

		for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
		{
			table.AddInteger( lands_in_starting_hand );

			CProbabilityDistribution state( 40, 17, starting_hand_size, lands_in_starting_hand );
			state.DrawCards( 3 );
			table.AddProbability( state.TakeMeasurement( competitive_start_measurement ) );
			state.DrawCard();
			table.AddProbability( state.TakeMeasurement( competitive_start_measurement ) );
		}

		table.Print();
	}

	{
		CCalculateMulliganRulesResults results = CalculateMulliganRulesIgnoringColor( Play, 4, competitive_start_measurement );
		results.PrintEntireCalculation( "Mulligan rules calculation on the play with standard competitive start defintion." );
		results.PrintComplexTable( "Mulligan rules calculation on the play with standard competitive start defintion." );
	}
	
	{
		CProbabilityDistribution state_play( 40, 17 );
		CCalculateMulliganRulesResults mulligan_play = CalculateMulliganRulesIgnoringColor( Play, 4, competitive_start_measurement );
		state_play.DrawStartingHandFromRules( mulligan_play.m_rules );

		CProbabilityDistribution state_draw( 40, 17 );
		CCalculateMulliganRulesResults mulligan_draw = CalculateMulliganRulesIgnoringColor( Draw, 4, competitive_start_measurement );
		state_draw.DrawStartingHandFromRules( mulligan_play.m_rules );

		CTable table_lands_in_play( "Average number of lands in play on a given turn, using mulligan m_rules with a standard competitive start definition." );
		table_lands_in_play.AddColumn( "turn" );
		table_lands_in_play.AddColumn( "play" );
		table_lands_in_play.AddColumn( "draw" );

		CTable table_land_count_difference( "Difference in average number of lands in play when on the draw versus on the play, using mulligan m_rules with a standard competitive start definition." );
		table_land_count_difference.AddColumn( "turn" );
		table_land_count_difference.AddColumn( "difference" );

		CTable table_growth_rate( "Rate of mana growth, using mulligan m_rules with a standard competitive start definition." );
		table_growth_rate.AddColumn( "turn" );
		table_growth_rate.AddColumn( "play" );
		table_growth_rate.AddColumn( "draw" );

		double previous_lands_in_play_play = 0.0;
		double previous_lands_in_play_draw = 0.0;

		for ( int turn = 1; turn < 15; ++turn )
		{
			table_lands_in_play.AddInteger( turn );
			table_land_count_difference.AddInteger( turn );
			table_growth_rate.AddInteger( turn );

			const double lands_in_play_play = state_play.TakeMeasurement(
				[ turn ]( const CProbabilityDistributionEntry& entry )
				{
					return std::min( turn, entry.m_player.m_hand.m_lands );
				}
			);
			table_lands_in_play.AddNumber( lands_in_play_play );
			state_play.DrawCard();

			state_draw.DrawCard();
			const double lands_in_play_draw = state_play.TakeMeasurement(
				[ turn ]( const CProbabilityDistributionEntry& entry )
				{
					return std::min( turn, entry.m_player.m_hand.m_lands );
				}
			);
			table_lands_in_play.AddNumber( lands_in_play_draw );

			table_land_count_difference.AddNumber( lands_in_play_draw - lands_in_play_play, 3 );

			table_growth_rate.AddNumber( lands_in_play_play - previous_lands_in_play_play, 3 );
			table_growth_rate.AddNumber( lands_in_play_draw - previous_lands_in_play_draw, 3 );

			previous_lands_in_play_play = lands_in_play_play;
			previous_lands_in_play_draw = lands_in_play_draw;
		}

		table_lands_in_play.Print();
		table_land_count_difference.Print();

		{
			CProbabilityDistribution state( 40, 17 );
			const double land_rate_before_game = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_library.GetLandRatio();
				}
			);

			double land_rate_after_mulligans_play;
			{
				CProbabilityDistribution state_play( state );
				CCalculateMulliganRulesResults mulligan_play = CalculateMulliganRulesIgnoringColor( Play, 4, competitive_start_measurement );
				state_play.DrawStartingHandFromRules( mulligan_play.m_rules );
				land_rate_after_mulligans_play = state_play.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_library.GetLandRatio();
					}
				);
			}

			double land_rate_after_mulligans_draw;
			{
				CProbabilityDistribution state_draw( state );
				CCalculateMulliganRulesResults mulligan_draw = CalculateMulliganRulesIgnoringColor( Draw, 4, competitive_start_measurement );
				state_draw.DrawStartingHandFromRules( mulligan_draw.m_rules );
				land_rate_after_mulligans_draw = state_draw.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_library.GetLandRatio();
					}
				);
			}

			CTable table( "Land rate of library before game and after mulligans, using mulligan m_rules with a standard competitive start definition." );
			table.AddAddendum( "before game" );
			table.AddProbability( land_rate_before_game );
			table.AddAddendum( "after mulligans" );
			table.AddProbability( ( land_rate_after_mulligans_play + land_rate_after_mulligans_draw ) / 2.0 );
			table.Print();
		}

		table_growth_rate.Print();
	}

	{
		auto entry_complete_condition = []( const CGameProbabilityDistributionEntry& entry )
		{
			return entry.m_user_data != 0.0;
		};
		auto entry_incomplete_condition = []( const CGameProbabilityDistributionEntry& entry )
		{
			return entry.m_user_data == 0.0;
		};

		CTable table( "Chance of the player on the play getting to a given land count before the player on the draw, using mulligan m_rules with a standard competitive start definition." );
		table.AddColumn( "land" );
		table.AddColumn( "play" );
		table.AddColumn( "draw" );
		for ( int converted_casting_cost = 1; converted_casting_cost < 18; ++converted_casting_cost )
		{
			CGameProbabilityDistribution state( 40, 17 );

			CCalculateMulliganRulesResults mulligan_play = CalculateMulliganRulesIgnoringColor( Play, 4, competitive_start_measurement );
			state.DrawStartingHandFromRules( mulligan_play.m_rules, 0 );

			CCalculateMulliganRulesResults mulligan_draw = CalculateMulliganRulesIgnoringColor( Draw, 4, competitive_start_measurement );
			state.DrawStartingHandFromRules( mulligan_draw.m_rules, 1 );

			int turn = 1;

			while ( true )
			{
				state.ForEach(
					[ turn, converted_casting_cost ]( CGameProbabilityDistributionEntry& entry )
					{
						bool player_on_play_can_cast = entry.m_players[ 0 ].m_hand.m_lands >= converted_casting_cost && turn >= converted_casting_cost;
						if ( player_on_play_can_cast && entry.m_user_data == 0.0 )
						{
							entry.m_user_data = turn;
						}
					}
				);

				state.DrawIf( entry_incomplete_condition, 1 );

				state.ForEach(
					[ turn, converted_casting_cost ]( CGameProbabilityDistributionEntry& entry )
					{
						bool player_on_draw_can_cast = entry.m_players[ 1 ].m_hand.m_lands >= converted_casting_cost && turn >= converted_casting_cost;
						if ( player_on_draw_can_cast && entry.m_user_data == 0.0 )
						{
							entry.m_user_data = -turn;
						}
					}
				);

				++turn;

				if ( state.TestIf( entry_complete_condition ) ) break;

				state.DrawIf( entry_incomplete_condition, 0 );
			}

			const double chance_player_on_the_play_is_first = state.TakeMeasurement(
				[]( const CGameProbabilityDistributionEntry& entry )
				{
					return entry.m_user_data > 0.0;
				}
			);
			const double chance_player_on_the_draw_is_first = state.TakeMeasurement(
				[]( const CGameProbabilityDistributionEntry& entry )
				{
					return entry.m_user_data < 0.0;
				}
			);

			table.AddInteger( converted_casting_cost );
			table.AddProbability( chance_player_on_the_play_is_first );
			table.AddProbability( chance_player_on_the_draw_is_first );
		}

		table.Print();
	}

	CalculateMulliganRulesIgnoringColor( Play, 4, competitive_start_measurement ).PrintCleanTable(
		"Mulligan rules on the play with a standard deck and a standard definition of competitive start."
	);

	CalculateMulliganRulesIgnoringColor( Draw, 4, competitive_start_measurement ).PrintCleanTable(
		"Mulligan rules on the draw with a standard deck and a standard definition of competitive start."
	);

	CalculateMulliganRulesIgnoringColor( Play, 4, competitive_start_measurement, 40, 19 ).PrintCleanTable(
		"Mulligan rules on the play with a 19-land deck and a standard definition of competitive start."
	);

	CalculateMulliganRulesIgnoringColor( Draw, 4, competitive_start_measurement, 40, 19 ).PrintCleanTable(
		"Mulligan rules on the draw with a 19-land deck and a standard definition of competitive start."
	);

	{
		auto low_curve_competitive_start_measurement = []( const CProbabilityDistributionEntry& entry )
		{
			return entry.m_player.m_hand.m_lands >= 3 && entry.m_player.m_hand.m_spells >= 5;
		};

		CalculateMulliganRulesIgnoringColor( Play, 4, low_curve_competitive_start_measurement ).PrintCleanTable(
			"Mulligan rules on the play with a standard deck and a low-curve definition of competitive start."
		);

		CalculateMulliganRulesIgnoringColor( Draw, 4, low_curve_competitive_start_measurement ).PrintCleanTable(
			"Mulligan rules on the draw with a standard deck and a low-curve definition of competitive start."
		);
	}
}