#include "magic_probability_toolkit.h"
#include "table_output.h"
#include "mulligan_rules.h"
#include "turn_cost.h"

using namespace mpt;
using namespace tbo;
using namespace mlr;
using namespace ctc;

void article_13()
{
	const int casting_cost_table[] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };

	{
		CTable table( "Turn cost." );
		table.AddColumn( "mana cost" );
		table.AddColumn( "turn cost" );

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry ) {};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry )
		{
			return entry.m_player.m_hand.m_lands >= 4 && entry.m_player.m_hand.m_spells >= 4;
		};

		CProbabilityDistribution::TPileType deck;
		deck.m_cards[ CProbabilityDistribution::TCardSetType::Land ] = 17;
		deck.m_cards[ CProbabilityDistribution::TCardSetType::Spell ] = 23;

		CMulliganRules< CProbabilityDistribution::TCardSetType > rules_play = CalculateMulliganRules<
			CProbabilityDistribution,
			decltype( initial_hand_evaluation ),
			decltype( competitive_start_measurement )
		>( Play, 4, initial_hand_evaluation, competitive_start_measurement, deck );

		CMulliganRules< CProbabilityDistribution::TCardSetType > rules_draw = CalculateMulliganRules<
			CProbabilityDistribution,
			decltype( initial_hand_evaluation ),
			decltype( competitive_start_measurement )
		>( Draw, 4, initial_hand_evaluation, competitive_start_measurement, deck );

		for ( int casting_cost_index = 0; casting_cost_index < std::extent< decltype( casting_cost_table ) >::value; ++casting_cost_index )
		{
			const int casting_cost = casting_cost_table[ casting_cost_index ];

			table.AddInteger( casting_cost );

			double average_turn_cost = 0.0;

			{
				CProbabilityDistribution state( deck );
				state.DrawStartingHandFromRules( rules_play );

				state.ForEach(
					[]( CProbabilityDistributionEntry& entry )
					{
						entry.m_user_data = entry.m_player.m_hand.GetSize();
					}
				);
				state.DrawUntil(
					[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size + 1;
						return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
					}
				);
				const double turn_cost = state.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry ) -> int
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size + 1;
						return turn;
					}
				);

				average_turn_cost += turn_cost;
			}

			{
				CProbabilityDistribution state( deck );
				state.DrawStartingHandFromRules( rules_draw );

				state.ForEach(
					[]( CProbabilityDistributionEntry& entry )
					{
						entry.m_user_data = entry.m_player.m_hand.GetSize();
					}
				);
				state.DrawCard();
				state.DrawUntil(
					[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
						return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
					}
				);
				const double turn_cost = state.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry ) -> int
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
						return turn;
					}
				);

				average_turn_cost += turn_cost;
			}
			
			table.AddNumber( average_turn_cost / 2.0 );
		}

		table.Print();
	}

	{
		CTable table( "Turn cost with various land counts." );
		table.AddColumn( "mana cost" );

		const int land_count_table[] = { 16, 17, 18, 19, 20 };

		for ( int land_count_index = 0; land_count_index < std::extent< decltype( land_count_table ) >::value; ++land_count_index )
		{
			const int land_count = land_count_table[ land_count_index ];
			table.AddColumn( "  %d", land_count );
		}

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry ) {};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry )
		{
			return entry.m_player.m_hand.m_lands >= 4 && entry.m_player.m_hand.m_spells >= 4;
		};

		for ( int casting_cost_index = 0; casting_cost_index < std::extent< decltype( casting_cost_table ) >::value; ++casting_cost_index )
		{
			const int casting_cost = casting_cost_table[ casting_cost_index ];

			table.AddInteger( casting_cost );

			for ( int land_count_index = 0; land_count_index < std::extent< decltype( land_count_table ) >::value; ++land_count_index )
			{
				const int land_count = land_count_table[ land_count_index ];

				double average_turn_cost = 0.0;

				CProbabilityDistribution::TPileType deck;
				deck.m_cards[ CProbabilityDistribution::TCardSetType::Land ] = land_count;
				deck.m_cards[ CProbabilityDistribution::TCardSetType::Spell ] = 40 - land_count;

				{
					CMulliganRules< CProbabilityDistribution::TCardSetType > rules = CalculateMulliganRules<
						CProbabilityDistribution,
						decltype( initial_hand_evaluation ),
						decltype( competitive_start_measurement )
					>( Play, 4, initial_hand_evaluation, competitive_start_measurement, deck );

					CProbabilityDistribution state( deck );
					state.DrawStartingHandFromRules( rules );

					state.ForEach(
						[]( CProbabilityDistributionEntry& entry )
						{
							entry.m_user_data = entry.m_player.m_hand.GetSize();
						}
					);
					state.DrawUntil(
						[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
						{
							const int starting_hand_size = int( entry.m_user_data );
							const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size + 1;
							return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
						}
					);
					const double turn_cost = state.TakeMeasurement(
						[]( const CProbabilityDistributionEntry& entry ) -> int
						{
							const int starting_hand_size = int( entry.m_user_data );
							const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size + 1;
							return turn;
						}
					);

					average_turn_cost += turn_cost;
				}

				{
					CMulliganRules< CProbabilityDistribution::TCardSetType > rules = CalculateMulliganRules<
						CProbabilityDistribution,
						decltype( initial_hand_evaluation ),
						decltype( competitive_start_measurement )
					>( Draw, 4, initial_hand_evaluation, competitive_start_measurement, deck );

					CProbabilityDistribution state( deck );
					state.DrawStartingHandFromRules( rules );

					state.ForEach(
						[]( CProbabilityDistributionEntry& entry )
						{
							entry.m_user_data = entry.m_player.m_hand.GetSize();
						}
					);
					state.DrawCard();
					state.DrawUntil(
						[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
						{
							const int starting_hand_size = int( entry.m_user_data );
							const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
							return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
						}
					);
					const double turn_cost = state.TakeMeasurement(
						[]( const CProbabilityDistributionEntry& entry ) -> int
						{
							const int starting_hand_size = int( entry.m_user_data );
							const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
							return turn;
						}
					);

					average_turn_cost += turn_cost;
				}

				average_turn_cost /= 2.0;
			
				table.AddNumber( average_turn_cost, 2 );
			}
		}

		table.Print();
	}

	{
		CTable table( "Drop in turn cost of 6-mana spell by adding land." );
		table.AddColumn( "land added" );
		table.AddColumn( "drop in turn cost" );

		const int land_count_table[] = { 16, 17, 18, 19, 20 };

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry ) {};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry )
		{
			return entry.m_player.m_hand.m_lands >= 4 && entry.m_player.m_hand.m_spells >= 4;
		};

		const int casting_cost = 6;

		double previous_turn_cost = 0.0;

		for ( int land_count_index = 0; land_count_index < std::extent< decltype( land_count_table ) >::value; ++land_count_index )
		{
			const int land_count = land_count_table[ land_count_index ];

			if ( land_count_index > 0 )
			{
				table.AddInteger( land_count );
			}

			double average_turn_cost = 0.0;

			CProbabilityDistribution::TPileType deck;
			deck.m_cards[ CProbabilityDistribution::TCardSetType::Land ] = land_count;
			deck.m_cards[ CProbabilityDistribution::TCardSetType::Spell ] = 40 - land_count;

			{
				CMulliganRules< CProbabilityDistribution::TCardSetType > rules = CalculateMulliganRules<
					CProbabilityDistribution,
					decltype( initial_hand_evaluation ),
					decltype( competitive_start_measurement )
				>( Play, 4, initial_hand_evaluation, competitive_start_measurement, deck );

				CProbabilityDistribution state( deck );
				state.DrawStartingHandFromRules( rules );

				state.ForEach(
					[]( CProbabilityDistributionEntry& entry )
					{
						entry.m_user_data = entry.m_player.m_hand.GetSize();
					}
				);
				state.DrawUntil(
					[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size + 1;
						return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
					}
				);
				const double turn_cost = state.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry ) -> int
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size + 1;
						return turn;
					}
				);

				average_turn_cost += turn_cost;
			}

			{
				CMulliganRules< CProbabilityDistribution::TCardSetType > rules = CalculateMulliganRules<
					CProbabilityDistribution,
					decltype( initial_hand_evaluation ),
					decltype( competitive_start_measurement )
				>( Draw, 4, initial_hand_evaluation, competitive_start_measurement, deck );

				CProbabilityDistribution state( deck );
				state.DrawStartingHandFromRules( rules );

				state.ForEach(
					[]( CProbabilityDistributionEntry& entry )
					{
						entry.m_user_data = entry.m_player.m_hand.GetSize();
					}
				);
				state.DrawCard();
				state.DrawUntil(
					[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> bool
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
						return std::min( turn, entry.m_player.m_hand.m_lands ) >= casting_cost;
					}
				);
				const double turn_cost = state.TakeMeasurement(
					[]( const CProbabilityDistributionEntry& entry ) -> int
					{
						const int starting_hand_size = int( entry.m_user_data );
						const int turn = entry.m_player.m_hand.GetSize() - starting_hand_size;
						return turn;
					}
				);

				average_turn_cost += turn_cost;
			}

			average_turn_cost /= 2.0;
			
			if ( land_count_index > 0 )
			{
				table.AddNumber( previous_turn_cost - average_turn_cost, 2 );
			}

			previous_turn_cost = average_turn_cost;
		}

		table.Print();
	}

	const int maximum_number_of_special_cards = 2;

	{
		typedef CardSet::IgnoringColorWithRamp Card;
		typedef CardSetTypes< Card >::CPile CPile;
		typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
		typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

		CTable table_turn_cost( "Turn cost with elves, with a total of 18 lands and elves in the deck.  Elves count as mana sources in mulligans." );
		table_turn_cost.AddColumn( "mana cost" );
		for ( int ramp_count = 0; ramp_count <= maximum_number_of_special_cards; ++ramp_count )
		{
			table_turn_cost.AddColumn( "%d elves", ramp_count );
		}

		CTable table_difference( "Difference in turn cost between two elves and zero elves." );
		table_difference.AddColumn( "mana cost" );
		table_difference.AddColumn( "difference" );

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry )
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;
			if ( entry.m_player.m_hand.m_cards[ TCardSetType::Land ] < 1 )
			{
				entry.m_user_data = 1.0;
			}
		};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;

			if ( entry.m_user_data != 0.0 ) return false;

			const int mana_sources = entry.m_player.m_hand.m_cards[ TCardSetType::Land ] + entry.m_player.m_hand.m_cards[ TCardSetType::Ramp ];
			return mana_sources >= 4 && entry.m_player.m_hand.m_cards[ TCardSetType::Spell ] >= 4;
		};

		for ( int casting_cost_index = 0; casting_cost_index < std::extent< decltype( casting_cost_table ) >::value; ++casting_cost_index )
		{
			const int casting_cost = casting_cost_table[ casting_cost_index ];

			double zero_ramp_turn_cost;

			table_turn_cost.AddInteger( casting_cost );
			table_difference.AddInteger( casting_cost );

			auto execute_turn = [ casting_cost ]( CProbabilityDistributionEntry& entry ) -> bool
			{
				typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

				const int casting_cost_of_ramp = 1;

				const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];

				const int land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
				int mana = land_in_play + entry.m_player.m_hand.m_cards[ Card::RampInPlay ];

				if ( mana >= casting_cost )
				{
					return true;
				}

				if ( entry.m_player.m_hand.m_cards[ Card::Ramp ] > 0 )
				{
					const int ramp_cast_count = std::min( mana / casting_cost_of_ramp, entry.m_player.m_hand.m_cards[ Card::Ramp ] );
					entry.m_player.m_hand.m_cards[ Card::Ramp ] -= ramp_cast_count;
					entry.m_player.m_hand.m_cards[ Card::RampInPlay ] += ramp_cast_count;
				}

				++entry.m_player.m_hand.m_cards[ Card::Turn ];
				return false;
			};

			for ( int ramp_count = 0; ramp_count <= maximum_number_of_special_cards; ++ramp_count )
			{
				CPile deck;
				deck.m_cards[ Card::Land ] = 18 - ramp_count;
				deck.m_cards[ Card::Ramp ] = ramp_count;
				deck.m_cards[ Card::Spell ] = 22;
				deck.m_cards[ Card::RampInPlay ] = 0;
				deck.m_cards[ Card::Turn ] = 0;

				const double average_turn_cost = ctc::CalculateTurnCost<
					CProbabilityDistribution,
					decltype( initial_hand_evaluation ),
					decltype( competitive_start_measurement ),
					decltype( execute_turn )
				>( deck, initial_hand_evaluation, competitive_start_measurement, execute_turn );

				table_turn_cost.AddNumber( average_turn_cost, 2 );

				if ( ramp_count == 0 )
				{
					zero_ramp_turn_cost = average_turn_cost;
				}
				else if ( ramp_count == maximum_number_of_special_cards )
				{
					table_difference.AddNumber( average_turn_cost - zero_ramp_turn_cost, 2 );
				}
			}
		}

		table_turn_cost.Print();
		table_difference.Print();
	}

	{
		typedef CardSet::IgnoringColorWithRamp Card;
		typedef CardSetTypes< Card >::CPile CPile;
		typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
		typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

		CTable table_turn_cost( "Turn cost with keyrunes, with a total of 18 lands and keyrunes in the deck." );
		table_turn_cost.AddColumn( "mana cost" );
		for ( int ramp_count = 0; ramp_count <= maximum_number_of_special_cards; ++ramp_count )
		{
			table_turn_cost.AddColumn( "%d keyrunes", ramp_count );
		}

		CTable table_difference( "Difference in turn cost between two keyrunes and zero keyrunes." );
		table_difference.AddColumn( "mana cost" );
		table_difference.AddColumn( "difference" );

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry ) {};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;

			const int spells = entry.m_player.m_hand.m_cards[ TCardSetType::Spell ] + entry.m_player.m_hand.m_cards[ TCardSetType::Ramp ];
			return entry.m_player.m_hand.m_cards[ TCardSetType::Land ] >= 4 && spells >= 4;
		};

		for ( int casting_cost_index = 0; casting_cost_index < std::extent< decltype( casting_cost_table ) >::value; ++casting_cost_index )
		{
			const int casting_cost = casting_cost_table[ casting_cost_index ];

			double zero_ramp_turn_cost;

			table_turn_cost.AddInteger( casting_cost );
			table_difference.AddInteger( casting_cost );

			auto execute_turn = [ casting_cost ]( CProbabilityDistributionEntry& entry ) -> bool
			{
				typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

				const int casting_cost_of_ramp = 3;

				const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];

				const int land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
				int mana = land_in_play + entry.m_player.m_hand.m_cards[ Card::RampInPlay ];

				if ( mana >= casting_cost )
				{
					return true;
				}

				if ( entry.m_player.m_hand.m_cards[ Card::Ramp ] > 0 )
				{
					const int ramp_cast_count = std::min( mana / casting_cost_of_ramp, entry.m_player.m_hand.m_cards[ Card::Ramp ] );
					entry.m_player.m_hand.m_cards[ Card::Ramp ] -= ramp_cast_count;
					entry.m_player.m_hand.m_cards[ Card::RampInPlay ] += ramp_cast_count;
				}

				++entry.m_player.m_hand.m_cards[ Card::Turn ];
				return false;
			};

			for ( int ramp_count = 0; ramp_count <= maximum_number_of_special_cards; ++ramp_count )
			{
				CPile deck;
				deck.m_cards[ Card::Land ] = 18 - ramp_count;
				deck.m_cards[ Card::Ramp ] = ramp_count;
				deck.m_cards[ Card::Spell ] = 22;
				deck.m_cards[ Card::RampInPlay ] = 0;
				deck.m_cards[ Card::Turn ] = 0;

				const double average_turn_cost = ctc::CalculateTurnCost<
					CProbabilityDistribution,
					decltype( initial_hand_evaluation ),
					decltype( competitive_start_measurement ),
					decltype( execute_turn )
				>( deck, initial_hand_evaluation, competitive_start_measurement, execute_turn );

				table_turn_cost.AddNumber( average_turn_cost, 2 );

				if ( ramp_count == 0 )
				{
					zero_ramp_turn_cost = average_turn_cost;
				}
				else if ( ramp_count == maximum_number_of_special_cards )
				{
					table_difference.AddNumber( average_turn_cost - zero_ramp_turn_cost, 2 );
				}
			}
		}

		table_turn_cost.Print();
		table_difference.Print();
	}

	{
		typedef CardSet::IgnoringColorWithRamp Card;
		typedef CardSetTypes< Card >::CPile CPile;
		typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
		typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

		CTable table_turn_cost( "Turn cost with Seek the Horizon." );
		table_turn_cost.AddColumn( "mana cost" );
		for ( int ramp_count = 0; ramp_count <= maximum_number_of_special_cards; ++ramp_count )
		{
			table_turn_cost.AddColumn( "%d copies", ramp_count );
		}

		CTable table_difference( "Difference in turn cost between two copies and zero copies." );
		table_difference.AddColumn( "mana cost" );
		table_difference.AddColumn( "difference" );

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry ) {};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

			const int spells = entry.m_player.m_hand.m_cards[ Card::Spell ] + entry.m_player.m_hand.m_cards[ Card::Ramp ];
			return entry.m_player.m_hand.m_cards[ Card::Land ] >= 4 && spells >= 4;
		};

		for ( int casting_cost_index = 0; casting_cost_index < std::extent< decltype( casting_cost_table ) >::value; ++casting_cost_index )
		{
			const int casting_cost = casting_cost_table[ casting_cost_index ];

			double zero_ramp_turn_cost;

			table_turn_cost.AddInteger( casting_cost );
			table_difference.AddInteger( casting_cost );

			auto execute_turn = [ casting_cost ]( CProbabilityDistributionEntry& entry ) -> bool
			{
				typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

				const int casting_cost_of_ramp = 4;
				const int lands_fetched_by_ramp = 3;

				const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];

				const int land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
				int mana = land_in_play;

				if ( mana >= casting_cost )
				{
					return true;
				}

				if ( entry.m_player.m_hand.m_cards[ Card::Ramp ] > 0 )
				{
					const int ramp_cast_count = std::min( mana / casting_cost_of_ramp, entry.m_player.m_hand.m_cards[ Card::Ramp ] );
					entry.m_player.m_hand.m_cards[ Card::Ramp ] -= ramp_cast_count;

					const int lands_fetched = std::min( lands_fetched_by_ramp * ramp_cast_count, entry.m_player.m_library.m_cards[ Card::Land ] );
					entry.m_player.m_hand.m_cards[ Card::Land ] += lands_fetched;
					entry.m_player.m_library.m_cards[ Card::Land ] -= lands_fetched;
				}

				++entry.m_player.m_hand.m_cards[ Card::Turn ];
				return false;
			};

			for ( int ramp_count = 0; ramp_count <= maximum_number_of_special_cards; ++ramp_count )
			{
				CPile deck;
				deck.m_cards[ Card::Land ] = 17;
				deck.m_cards[ Card::Ramp ] = ramp_count;
				deck.m_cards[ Card::Spell ] = 23 - ramp_count;
				deck.m_cards[ Card::RampInPlay ] = 0;
				deck.m_cards[ Card::Turn ] = 0;

				const double average_turn_cost = ctc::CalculateTurnCost<
					CProbabilityDistribution,
					decltype( initial_hand_evaluation ),
					decltype( competitive_start_measurement ),
					decltype( execute_turn )
				>( deck, initial_hand_evaluation, competitive_start_measurement, execute_turn );

				table_turn_cost.AddNumber( average_turn_cost, 2 );

				if ( ramp_count == 0 )
				{
					zero_ramp_turn_cost = average_turn_cost;
				}
				else if ( ramp_count == maximum_number_of_special_cards )
				{
					table_difference.AddNumber( average_turn_cost - zero_ramp_turn_cost, 2 );
				}
			}
		}

		table_turn_cost.Print();
		table_difference.Print();
	}

	{
		typedef CardSet::IgnoringColorWithCantrips Card;
		typedef CardSetTypes< Card >::CPile CPile;
		typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
		typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

		CTable table_turn_cost( "Turn cost with cantrips." );
		table_turn_cost.AddColumn( "mana cost" );
		for ( int cantrip_count = 0; cantrip_count <= maximum_number_of_special_cards; ++cantrip_count )
		{
			table_turn_cost.AddColumn( "%d cantrips", cantrip_count );
		}

		CTable table_difference( "Difference in turn cost between two cantrips and zero cantrips." );
		table_difference.AddColumn( "mana cost" );
		table_difference.AddColumn( "difference" );

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry ) {};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

			const int spells = entry.m_player.m_hand.m_cards[ Card::Spell ] + entry.m_player.m_hand.m_cards[ Card::Cantrip ];
			return entry.m_player.m_hand.m_cards[ Card::Land ] >= 4 && spells >= 4;
		};

		auto get_turn = []( const CProbabilityDistributionEntry& entry ) -> int
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;
			return entry.m_player.m_hand.m_cards[ Card::Turn ];
		};

		auto set_first_turn = []( CProbabilityDistributionEntry& entry )
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;
			entry.m_player.m_hand.m_cards[ Card::Turn ] = 1;
		};

		for ( int casting_cost_index = 0; casting_cost_index < std::extent< decltype( casting_cost_table ) >::value; ++casting_cost_index )
		{
			const int casting_cost = casting_cost_table[ casting_cost_index ];

			double zero_cantrip_turn_cost;

			table_turn_cost.AddInteger( casting_cost );
			table_difference.AddInteger( casting_cost );

			auto execute_turn = [ casting_cost ]( CProbabilityDistributionEntry& entry ) -> bool
			{
				typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

				const int casting_cost_of_cantrip = 1;

				const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];

				const int land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
				int mana = land_in_play - entry.m_player.m_hand.m_cards[ Card::ManaSpent ];

				if ( mana >= casting_cost )
				{
					return true;
				}

				if ( entry.m_player.m_hand.m_cards[ Card::Cantrip ] > 0 )
				{
					const int cantrip_cast_count = std::min( mana / casting_cost_of_cantrip, entry.m_player.m_hand.m_cards[ Card::Cantrip ] );
					entry.m_player.m_hand.m_cards[ Card::Cantrip ] -= cantrip_cast_count;
					entry.m_player.m_hand.m_cards[ Card::CardsToDraw ] += cantrip_cast_count;
					entry.m_player.m_hand.m_cards[ Card::ManaSpent ] += casting_cost_of_cantrip * cantrip_cast_count;
				}

				if ( entry.m_player.m_hand.m_cards[ Card::CardsToDraw ] > 0 )
				{
					--entry.m_player.m_hand.m_cards[ Card::CardsToDraw ];
					return false;
				}

				++entry.m_player.m_hand.m_cards[ Card::Turn ];
				entry.m_player.m_hand.m_cards[ Card::ManaSpent ] = 0;
				return false;
			};

			for ( int cantrip_count = 0; cantrip_count <= maximum_number_of_special_cards; ++cantrip_count )
			{
				CPile deck;
				deck.m_cards[ Card::Land ] = 17;
				deck.m_cards[ Card::Cantrip ] = cantrip_count;
				deck.m_cards[ Card::Spell ] = 23 - cantrip_count;
				deck.m_cards[ Card::Turn ] = 0;
				deck.m_cards[ Card::CardsToDraw ] = 0;
				deck.m_cards[ Card::ManaSpent ] = 0;

				const double average_turn_cost = ctc::CalculateTurnCost<
					CProbabilityDistribution,
					decltype( initial_hand_evaluation ),
					decltype( competitive_start_measurement ),
					decltype( execute_turn )
				>( deck, initial_hand_evaluation, competitive_start_measurement, execute_turn );

				table_turn_cost.AddNumber( average_turn_cost, 2 );

				if ( cantrip_count == 0 )
				{
					zero_cantrip_turn_cost = average_turn_cost;
				}
				else if ( cantrip_count == maximum_number_of_special_cards )
				{
					table_difference.AddNumber( average_turn_cost - zero_cantrip_turn_cost, 2 );
				}
			}
		}

		table_turn_cost.Print();
		table_difference.Print();
	}

	{
		typedef CardSet::IgnoringColorWithRamp Card;
		typedef CardSetTypes< Card >::CPile CPile;
		typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
		typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

		CTable table_turn_cost( "Turn cost with moxen, with a total of 17 lands and moxen in the deck.  Moxen count as mana sources in mulligans." );
		table_turn_cost.AddColumn( "mana cost" );
		for ( int ramp_count = 0; ramp_count <= maximum_number_of_special_cards; ++ramp_count )
		{
			table_turn_cost.AddColumn( "%d moxen", ramp_count );
		}

		CTable table_difference( "Difference in turn cost between two moxen and zero moxen." );
		table_difference.AddColumn( "mana cost" );
		table_difference.AddColumn( "difference" );

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry ) {};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;

			const int mana_sources = entry.m_player.m_hand.m_cards[ TCardSetType::Land ] + entry.m_player.m_hand.m_cards[ TCardSetType::Ramp ];
			return mana_sources >= 4 && entry.m_player.m_hand.m_cards[ TCardSetType::Spell ] >= 4;
		};

		for ( int casting_cost_index = 0; casting_cost_index < std::extent< decltype( casting_cost_table ) >::value; ++casting_cost_index )
		{
			const int casting_cost = casting_cost_table[ casting_cost_index ];

			double zero_ramp_turn_cost;

			table_turn_cost.AddInteger( casting_cost );
			table_difference.AddInteger( casting_cost );

			auto execute_turn = [ casting_cost ]( CProbabilityDistributionEntry& entry ) -> bool
			{
				typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

				if ( entry.m_player.m_hand.m_cards[ Card::Ramp ] > 0 )
				{
					const int ramp_cast_count = entry.m_player.m_hand.m_cards[ Card::Ramp ];
					entry.m_player.m_hand.m_cards[ Card::Ramp ] -= ramp_cast_count;
					entry.m_player.m_hand.m_cards[ Card::RampInPlay ] += ramp_cast_count;
				}

				const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];

				const int land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
				int mana = land_in_play + entry.m_player.m_hand.m_cards[ Card::RampInPlay ];

				if ( mana >= casting_cost )
				{
					return true;
				}

				++entry.m_player.m_hand.m_cards[ Card::Turn ];
				return false;
			};

			for ( int ramp_count = 0; ramp_count <= maximum_number_of_special_cards; ++ramp_count )
			{
				CPile deck;
				deck.m_cards[ Card::Land ] = 17 - ramp_count;
				deck.m_cards[ Card::Ramp ] = ramp_count;
				deck.m_cards[ Card::Spell ] = 23;
				deck.m_cards[ Card::RampInPlay ] = 0;
				deck.m_cards[ Card::Turn ] = 0;

				const double average_turn_cost = ctc::CalculateTurnCost<
					CProbabilityDistribution,
					decltype( initial_hand_evaluation ),
					decltype( competitive_start_measurement ),
					decltype( execute_turn )
				>( deck, initial_hand_evaluation, competitive_start_measurement, execute_turn );

				table_turn_cost.AddNumber( average_turn_cost, 2 );

				if ( ramp_count == 0 )
				{
					zero_ramp_turn_cost = average_turn_cost;
				}
				else if ( ramp_count == maximum_number_of_special_cards )
				{
					table_difference.AddNumber( average_turn_cost - zero_ramp_turn_cost, 2 );
				}
			}
		}

		table_turn_cost.Print();
		table_difference.Print();
	}
}