#include "magic_probability_toolkit.h"
#include "mulligan_rules.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

void article_12()
{
	auto mark_competitive_turn_three = []( CProbabilityDistributionEntry& entry )
	{
		entry.m_user_data = entry.m_player.m_hand.m_lands >= 3;
	};
	auto competitive_turn_five = []( const CProbabilityDistributionEntry& entry )
	{
		return entry.m_user_data && entry.m_player.m_hand.m_spells >= 5;
	};
	auto secondary_competitive_start_measurement = []( const CProbabilityDistributionEntry& entry )
	{
		return entry.m_player.m_hand.m_lands >= 3 && entry.m_player.m_hand.m_spells >= 3;
	};

	for ( int starting_hand_size = 7; starting_hand_size > 1; --starting_hand_size )
	{
		CTable table( "Chance of a competitive start for a low-curve deck with various land counts in a %d-card hand.", starting_hand_size );
		table.AddColumn( "land" );
		table.AddColumn( "spells" );
		table.AddColumn( "probability" );

		for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
		{
			CProbabilityDistribution state( 40, 17, starting_hand_size, lands_in_starting_hand );
			state.DrawCards( 3 );
			state.ForEach( mark_competitive_turn_three );
			state.DrawCards( 2 );
			const double chance_of_competitive_start = state.TakeMeasurement( competitive_turn_five );

			table.AddInteger( lands_in_starting_hand );
			table.AddInteger( starting_hand_size - lands_in_starting_hand );
			table.AddProbability( chance_of_competitive_start );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawCards( starting_hand_size );
			state.DrawCards( 3 );
			state.ForEach( mark_competitive_turn_three );
			state.DrawCards( 2 );
			const double chance_of_competitive_situation = state.TakeMeasurement( competitive_turn_five );

			table.AddAddendum( "overall chance of a competitive start" );
			table.AddProbability( chance_of_competitive_situation );
		}

		table.Print();
	}

	for ( int starting_hand_size = 7; starting_hand_size > 2; --starting_hand_size )
	{
		CTable table( "Chance of a competitive start for a low-curve deck with various land counts in a %d-card hand, on the play.", starting_hand_size );
		table.AddColumn( "land" );
		table.AddColumn( "spells" );
		table.AddColumn( "probability" );

		for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
		{
			CProbabilityDistribution state( 40, 17, starting_hand_size, lands_in_starting_hand );
			state.DrawCards( 2 );
			state.ForEach( mark_competitive_turn_three );
			state.DrawCards( 2 );
			const double chance_of_competitive_situation = state.TakeMeasurement( competitive_turn_five );

			table.AddInteger( lands_in_starting_hand );
			table.AddInteger( starting_hand_size - lands_in_starting_hand );
			table.AddProbability( chance_of_competitive_situation );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawCards( starting_hand_size );
			state.DrawCards( 2 );
			state.ForEach( mark_competitive_turn_three );
			state.DrawCards( 2 );
			const double chance_of_competitive_situation = state.TakeMeasurement( competitive_turn_five );

			table.AddAddendum( "overall chance of a competitive start" );
			table.AddProbability( chance_of_competitive_situation );
		}

		table.Print();
	}
	
	for ( int starting_hand_size = 4; starting_hand_size > 1; --starting_hand_size )
	{
		CTable table( "Chance of a secondary competitive start for a low-curve deck with various land counts in a %d-card hand, on the play.", starting_hand_size );
		table.AddColumn( "land" );
		table.AddColumn( "spells" );
		table.AddColumn( "probability" );

		for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
		{
			CProbabilityDistribution state( 40, 17, starting_hand_size, lands_in_starting_hand );
			state.DrawCards( 3 );

			const double chance_of_secondary_competitive_start = state.TakeMeasurement( secondary_competitive_start_measurement );
			table.AddInteger( lands_in_starting_hand );
			table.AddInteger( starting_hand_size - lands_in_starting_hand );
			table.AddProbability( chance_of_secondary_competitive_start );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawCards( starting_hand_size );
			state.DrawCards( 3 );
			const double chance_of_secondary_competitive_start = state.TakeMeasurement( secondary_competitive_start_measurement );

			table.AddAddendum( "overall chance of a competitive start" );
			table.AddProbability( chance_of_secondary_competitive_start );
		}

		table.Print();
	}

	{
		CTable table( "Overall chance of a competitive start for various situations for a low-curve deck." );
		table.AddColumn( "situation" );
		table.AddColumn( "probability" );

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawCards( 7 );
			state.DrawCards( 3 );
			state.ForEach( mark_competitive_turn_three );
			state.DrawCards( 2 );
			const double chance_of_competitive_situation = state.TakeMeasurement( competitive_turn_five );
			table.AddString( "straight 7 on the draw" );
			table.AddProbability( chance_of_competitive_situation );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawCards( 7 );
			state.MulliganIf( MulliganRules::FastDeck_OnTheDraw );
			state.DrawCards( 3 );
			state.ForEach( mark_competitive_turn_three );
			state.DrawCards( 2 );
			const double chance_of_competitive_situation = state.TakeMeasurement( competitive_turn_five );
			table.AddString( "after mulligans on the draw" );
			table.AddProbability( chance_of_competitive_situation );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawCards( 7 );
			state.DrawCards( 2 );
			state.ForEach( mark_competitive_turn_three );
			state.DrawCards( 2 );
			const double chance_of_competitive_situation = state.TakeMeasurement( competitive_turn_five );
			table.AddString( "straight 7 on the play" );
			table.AddProbability( chance_of_competitive_situation );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawCards( 7 );
			state.MulliganIf( MulliganRules::FastDeck_OnThePlay );
			state.DrawCards( 2 );
			state.ForEach( mark_competitive_turn_three );
			state.DrawCards( 2 );
			const double chance_of_competitive_situation = state.TakeMeasurement( competitive_turn_five );
			table.AddString( "after mulligans on the play" );
			table.AddProbability( chance_of_competitive_situation );
		}

		table.Print();
	}
}