#include "magic_probability_toolkit.h"
#include "mulligan_rules.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

void article_13()
{
	{
		CTable table( "Chance of missing turn-3 land drop using standard and low-curve mulligan rules." );
		table.AddColumn( "mulligan rule" );
		table.AddColumn( "probability" );

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			state.DrawCards( 3 );
			const double chance_of_turn_three_land_drop = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 3;
				}
			);
			table.AddString( "standard" );
			table.AddProbability( 1.0 - chance_of_turn_three_land_drop );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::FastDeck_OnTheDraw );
			state.DrawCards( 3 );
			const double chance_of_turn_three_land_drop = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 3;
				}
			);
			table.AddString( "low-curve" );
			table.AddProbability( 1.0 - chance_of_turn_three_land_drop );
		}

		table.Print();
	}

	{
		CTable table( "Chance of missing turn-6 land drop using standard and low-curve mulligan rules." );
		table.AddColumn( "mulligan rule" );
		table.AddColumn( "probability" );

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			state.DrawCards( 6 );
			const double chance_of_turn_three_land_drop = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 6;
				}
			);
			table.AddString( "standard" );
			table.AddProbability( 1.0 - chance_of_turn_three_land_drop );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::FastDeck_OnTheDraw );
			state.DrawCards( 6 );
			const double chance_of_turn_three_land_drop = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 6;
				}
			);
			table.AddString( "low-curve" );
			table.AddProbability( 1.0 - chance_of_turn_three_land_drop );
		}

		table.Print();
	}
	{
		CTable table( "Average turn on which 5th land is played." );
		table.AddColumn( "mulligan rule" );
		table.AddColumn( "probability" );

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			state.ForEach(
				[]( CProbabilityDistributionEntry& entry )
				{
					entry.m_user_data = entry.m_player.m_hand.GetSize();
				}
			);
			state.DrawUntil(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 5;
				}
			);
			const double average_turn = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry ) -> double
				{
					const int number_of_draws = entry.m_player.m_hand.GetSize() - entry.m_user_data;
					return std::max( number_of_draws, 5 );
				}
			);
			table.AddString( "standard" );
			table.AddNumber( average_turn );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::FastDeck_OnTheDraw );
			state.ForEach(
				[]( CProbabilityDistributionEntry& entry )
				{
					entry.m_user_data = entry.m_player.m_hand.GetSize();
				}
			);
			state.DrawUntil(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 5;
				}
			);
			const double average_turn = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry ) -> double
				{
					const int number_of_draws = entry.m_player.m_hand.GetSize() - entry.m_user_data;
					return std::max( number_of_draws, 5 );
				}
			);
			table.AddString( "low-curve" );
			table.AddNumber( average_turn );
		}

		table.Print();
	}

	{
		CTable table( "Average number of spells available on turn 5." );
		table.AddColumn( "mulligan rule" );
		table.AddColumn( "spells" );

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			state.DrawCards( 5 );
			const double average_number_of_spells = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells;
				}
			);
			table.AddString( "standard" );
			table.AddNumber( average_number_of_spells );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::FastDeck_OnTheDraw );
			state.DrawCards( 5 );
			const double average_number_of_spells = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells;
				}
			);
			table.AddString( "low-curve" );
			table.AddNumber( average_number_of_spells );
		}

		table.Print();
	}

	{
		CTable table( "Chance of less than 5 spells available on turn 5." );
		table.AddColumn( "mulligan rule" );
		table.AddColumn( "probability" );

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			state.DrawCards( 5 );
			const double probability_of_less_than_five_spells = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells >= 5;
				}
			);
			table.AddString( "standard" );
			table.AddProbability( 1.0 - probability_of_less_than_five_spells );
		}

		{
			CProbabilityDistribution state( 40, 17 );
			state.DrawStartingHand( MulliganRules::FastDeck_OnTheDraw );
			state.DrawCards( 5 );
			const double probability_of_less_than_five_spells = state.TakeMeasurement(
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells >= 5;
				}
			);
			table.AddString( "low-curve" );
			table.AddProbability( 1.0 - probability_of_less_than_five_spells );
		}

		table.Print();
	}
}