#include "magic_probability_toolkit.h"
#include "mulligan_rules.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

void article_14()
{
	{
		CTable table( "Probability of mana screw for different deck land counts using standard mulligan rules." );
		table.AddColumn( "land" );
		table.AddColumn( "probability" );
		for ( int deck_land_count = 14; deck_land_count < 21; ++deck_land_count )
		{
			CProbabilityDistribution state( 40, deck_land_count );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			const double probability_of_mana_screw = state.TakeMeasurement( []( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands < 3;
				}
			);

			table.AddInteger( deck_land_count );
			table.AddProbability( probability_of_mana_screw );
		}
		table.Print();
	}

	{
		CTable table( "Probability of competitive start for different deck land counts using standard mulligan rules and standard competitive start definition." );
		table.AddColumn( "land" );
		table.AddColumn( "probability" );
		for ( int deck_land_count = 14; deck_land_count < 21; ++deck_land_count )
		{
			CProbabilityDistribution state( 40, deck_land_count );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			state.DrawCards( 4 );
			const double probability_of_competitive_start = state.TakeMeasurement( []( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= 4 && entry.m_player.m_hand.m_spells >= 4;
				}
			);

			table.AddInteger( deck_land_count );
			table.AddProbability( probability_of_competitive_start );
		}
		table.Print();
	}

	{
		auto mark_competitive_turn_three = []( CProbabilityDistributionEntry& entry )
		{
			entry.m_user_data = entry.m_player.m_hand.m_lands >= 3;
		};
		auto competitive_turn_five = []( const CProbabilityDistributionEntry& entry )
		{
			return entry.m_user_data && entry.m_player.m_hand.m_spells >= 5;
		};

		CTable table( "Probability of competitive start for different deck land counts using low-curve mulligan rules and low-curve competitive start definition." );
		table.AddColumn( "land" );
		table.AddColumn( "probability" );
		for ( int deck_land_count = 14; deck_land_count < 21; ++deck_land_count )
		{
			CProbabilityDistribution state( 40, deck_land_count );
			state.DrawStartingHand( MulliganRules::FastDeck_OnTheDraw );
			state.DrawCards( 3 );
			state.ForEach( []( CProbabilityDistributionEntry& entry )
				{
					entry.m_user_data = entry.m_player.m_hand.m_lands >= 3;
				}
			);
			state.DrawCards( 2 );
			const double probability_of_competitive_start = state.TakeMeasurement( []( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_user_data && entry.m_player.m_hand.m_spells >= 5;
				}
			);

			table.AddInteger( deck_land_count );
			table.AddProbability( probability_of_competitive_start );
		}
		table.Print();
	}

	{
		CTable table( "Average number of spells in hand on turn 10 for different deck land counts using standard mulligan." );
		table.AddColumn( "land" );
		table.AddColumn( "spells in hand" );
		for ( int deck_land_count = 14; deck_land_count < 21; ++deck_land_count )
		{
			CProbabilityDistribution state( 40, deck_land_count );
			state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
			state.DrawCards( 10 );
			const double average_number_of_spells = state.TakeMeasurement( []( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells;
				}
			);

			table.AddInteger( deck_land_count );
			table.AddNumber( average_number_of_spells );
		}
		table.Print();
	}


}