#include "magic_probability_toolkit.h"
#include "mulligan_rules.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

void article_15()
{
	CProbabilityDistribution template_state( 40, 17 );
	template_state.DrawStartingHand( MulliganRules::StandardDeck_OnTheDraw );
	template_state.ForEach(
		[]( CProbabilityDistributionEntry& entry )
		{
			entry.m_user_data = entry.m_player.m_hand.GetSize();
		}
	);

	double previous_average_turn = 0.0;

	CTable table( "Casting turn of artifacts with various casting costs." );
	table.AddColumn( "casting cost" );
	table.AddColumn( "casting turn" );
	table.AddColumn( "delta" );
	for ( int casting_cost = 1; casting_cost < 11; ++casting_cost )
	{
		CProbabilityDistribution state( template_state );
		state.DrawUntil(
			[ casting_cost ]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands >= casting_cost;
			}
		);
		const double average_turn = state.TakeMeasurement(
			[ casting_cost ]( const CProbabilityDistributionEntry& entry ) -> double
			{
				const int number_of_draws = entry.m_player.m_hand.GetSize() - entry.m_user_data;
				return std::max( number_of_draws, casting_cost );
			}
		);

		table.AddInteger( casting_cost );
		table.AddNumber( average_turn );
		table.AddNumber( average_turn - previous_average_turn );

		previous_average_turn = average_turn;
	}
	table.Print();

	/*
	{
		CProbabilityDistribution state_on_the_play( 40, 17, 7, 3 );
		CProbabilityDistribution state_on_the_draw( 40, 17, 7, 3 );
	
		CTable table( "Chance of hitting your land drop on a given turn for players on the play versus players on the draw." );
		table.AddColumn( "turn" );
		table.AddColumn( "on the play" );
		table.AddColumn( "on the draw" );
		for ( int turn = 1; turn < 11; ++turn )
		{
			if ( turn > 1 )
			{
				state_on_the_play.DrawCard();
			}
			state_on_the_draw.DrawCard();

			if ( turn > 2 )
			{
				const double chance_of_hitting_land_drop_on_the_play = state_on_the_play.TakeMeasurement(
					[ turn ]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.m_lands >= turn;
					}
				);
				const double chance_of_hitting_land_drop_on_the_draw = state_on_the_draw.TakeMeasurement(
					[ turn ]( const CProbabilityDistributionEntry& entry )
					{
						return entry.m_player.m_hand.m_lands >= turn;
					}
				);

				table.AddInteger( turn );
				table.AddProbability( chance_of_hitting_land_drop_on_the_play );
				table.AddProbability( chance_of_hitting_land_drop_on_the_draw );
			}
		}
		table.Print();
	}
	
	{
		CProbabilityDistribution state_on_the_play( 40, 17, 7, 3 );
		CProbabilityDistribution state_on_the_draw( 40, 17, 7, 3 );

		CTable table( "Average number of lands in play on a given turn for players on the play versus players on the draw." );
		table.AddColumn( "turn" );
		table.AddColumn( "on the play" );
		table.AddColumn( "on the draw" );
		for ( int turn = 1; turn < 15; ++turn )
		{
			if ( turn > 1 )
			{
				state_on_the_play.DrawCard();
			}
			state_on_the_draw.DrawCard();

			if ( turn > 2 )
			{
				const double lands_in_play_on_the_play = state_on_the_play.TakeMeasurement(
					[ turn ]( const CProbabilityDistributionEntry& entry )
					{
						return std::min( turn, entry.m_player.m_hand.m_lands );
					}
				);
				const double lands_in_play_on_the_draw = state_on_the_draw.TakeMeasurement(
					[ turn ]( const CProbabilityDistributionEntry& entry )
					{
						return std::min( turn, entry.m_player.m_hand.m_lands );
					}
				);

				table.AddInteger( turn );
				table.AddNumber( lands_in_play_on_the_play );
				table.AddNumber( lands_in_play_on_the_draw );
			}
		}
		table.Print();
	}

	{
		CTable table( "Chance of the player on the play getting to a given land count before the player on the draw." );
		table.AddColumn( "land" );
		table.AddColumn( "on the play" );
		table.AddColumn( "on the draw" );
		for ( int converted_casting_cost = 3; converted_casting_cost < 18; ++converted_casting_cost )
		{
			const int starting_hand_size = 7;

			CGameProbabilityDistribution state( 40, 17, 7, 3 );
			state.AlternatingDrawUntil(
				[ starting_hand_size, converted_casting_cost ]( const CGameProbabilityDistributionEntry& entry )
				{
					return entry.m_players[ 0 ].m_hand.m_lands >= converted_casting_cost && entry.m_players[ 0 ].m_hand.GetSize() >= starting_hand_size + converted_casting_cost - 1
						|| entry.m_players[ 1 ].m_hand.m_lands >= converted_casting_cost && entry.m_players[ 1 ].m_hand.GetSize() >= starting_hand_size + converted_casting_cost;
				}
			);
			const double chance_player_on_the_play_is_first = state.TakeMeasurement(
				[ starting_hand_size, converted_casting_cost ]( const CGameProbabilityDistributionEntry& entry )
				{
					return entry.m_players[ 0 ].m_hand.m_lands >= converted_casting_cost && entry.m_players[ 0 ].m_hand.GetSize() >= starting_hand_size + converted_casting_cost - 1;
				}
			);
			const double chance_player_on_the_draw_is_first = state.TakeMeasurement(
				[ starting_hand_size, converted_casting_cost ]( const CGameProbabilityDistributionEntry& entry )
				{
					return entry.m_players[ 1 ].m_hand.m_lands >= converted_casting_cost && entry.m_players[ 1 ].m_hand.GetSize() >= starting_hand_size + converted_casting_cost;
				}
			);

			table.AddInteger( converted_casting_cost );
			table.AddProbability( chance_player_on_the_play_is_first );
			table.AddProbability( chance_player_on_the_draw_is_first );
		}

		table.Print();
	}
	*/


	/*
	for ( int starting_land_count = 2; starting_land_count < 5; ++starting_land_count )
	{
		CProbabilityDistribution state( 40, 17, 7, starting_land_count );
	
		CTable table( "Chance of hitting your land drop on a given turn with a %d-land starting hand.", starting_land_count );
		table.AddColumn( "turn" );
		table.AddColumn( "probability" );
		for ( int turn = 1; turn < 13; ++turn )
		{
			state.DrawCard();
			const double chance_of_hitting_land_drop = state.TakeMeasurement(
				[ turn ]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands >= turn;
				}
			);
			table.AddInteger( turn );
			table.AddProbability( chance_of_hitting_land_drop );
		}
		table.Print();
	}
	*/


	/*
	const int turns[] = { 4, 6 };
	for ( int turn_idx = 0; turn_idx < std::extent< decltype( turns ) >::value; ++turn_idx )
	{
		const int turn = turns[ turn_idx ];
		auto missing_land_drop_measurement = [ turn ]( const CProbabilityDistributionEntry& entry )
		{
			return entry.m_player.m_hand.m_lands < turn;
		};


		CTable table( "Chance of having missed a land drop by turn %d.", turn );
		table.AddColumn( "player" );
		table.AddColumn( "probability" );

		{
			CProbabilityDistribution state( 40, 17, 7, 3 );
			state.DrawCards( turn - 1 );
			const double chance_of_missing_land_drop = state.TakeMeasurement( missing_land_drop_measurement );
			table.AddString( "on the play" );
			table.AddProbability( chance_of_missing_land_drop );
		}

		{
			CProbabilityDistribution state( 40, 17, 7, 3 );
			state.DrawCards( turn );
			const double chance_of_missing_land_drop = state.TakeMeasurement( missing_land_drop_measurement );
			table.AddString( "on the draw" );
			table.AddProbability( chance_of_missing_land_drop );
		}

		table.Print();
	}

	{
		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry )
		{
			return entry.m_player.m_hand.m_lands >= 4 && entry.m_player.m_hand.m_spells >= 4;
		};

		//for ( int starting_hand_size = 7; starting_hand_size >= 6; --starting_hand_size )
		{
			CTable table( "Chance of a competitive start with various land counts for both players." );
			table.AddColumn( "land" );
			table.AddColumn( "on the play" );
			table.AddColumn( "on the draw" );

			for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= 7; ++lands_in_starting_hand )
			{
				CProbabilityDistribution state( 40, 17, 7, lands_in_starting_hand );
				state.DrawCards( 3 );
				const double chance_of_competitive_start = state.TakeMeasurement( competitive_start_measurement );
				state.DrawCards( 1 );

				const double chance_of_competitive_start_ = state.TakeMeasurement( competitive_start_measurement );
				table.AddInteger( lands_in_starting_hand );
				table.AddProbability( chance_of_competitive_start );
				table.AddProbability( chance_of_competitive_start_ );
			}
		
			table.Print();
		}
	}
	*/

	/*
	CProbabilityDistribution no_mull_state( 40, 17 );
	no_mull_state.DrawCards( 7 );

	CProbabilityDistribution mull_state( 40, 17 );
	mull_state.DrawCards( 7 );
	mull_state.MulliganIf(
		[]( const CProbabilityDistributionEntry& entry ) -> bool
		{
			if ( entry.m_player.m_hand.GetSize() == 7 )
			{
				return entry.m_player.m_hand.m_lands < 3 || entry.m_player.m_hand.m_lands > 5;
			}
			if ( entry.m_player.m_hand.GetSize() == 6 )
			{
				return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 5;
			}
			return false;
		}
	);

	{
		double no_mull_land_table[ 8 ];
		double mull_land_table[ 8 ];
		no_mull_state.BinDistribution( no_mull_land_table, std::extent< decltype( no_mull_land_table ) >::value,
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands;
			}
		);
		mull_state.BinDistribution( mull_land_table, std::extent< decltype( mull_land_table ) >::value,
			[]( const CProbabilityDistributionEntry& entry )
			{
				return entry.m_player.m_hand.m_lands;
			}
		);
		CTable table( "Land count distribution with and without mulliganing." );
		table.AddColumn( "land" );
		table.AddColumn( "no mulligan" );
		table.AddColumn( "mulligan" );
		for ( int lands = 0; lands < 8; ++lands )
		{
			table.AddInteger( lands );
			table.AddProbability( no_mull_land_table[ lands ] );
			table.AddProbability( mull_land_table[ lands ] );
		}
		table.Print();
	}

	{
		const int prefered_lower_bound = 3;
		const int prefered_upper_bound = 5;

		double no_mull_land_situation_table[ 3 ];
		double mull_land_situation_table[ 3 ];
		no_mull_state.BinDistribution( no_mull_land_situation_table, std::extent< decltype( no_mull_land_situation_table ) >::value,
			[ prefered_lower_bound, prefered_upper_bound ]( const CProbabilityDistributionEntry& entry ) -> int
			{
				if ( entry.m_player.m_hand.m_lands < prefered_lower_bound ) return 0;
				if ( entry.m_player.m_hand.m_lands > prefered_upper_bound ) return 2;
				return 1;
			}
		);
		mull_state.BinDistribution( mull_land_situation_table, std::extent< decltype( mull_land_situation_table ) >::value,
			[ prefered_lower_bound, prefered_upper_bound ]( const CProbabilityDistributionEntry& entry ) -> int
			{
				if ( entry.m_player.m_hand.m_lands < prefered_lower_bound ) return 0;
				if ( entry.m_player.m_hand.m_lands > prefered_upper_bound ) return 2;
				return 1;
			}
		);

		CTable table( "Land situation distribution with and without mulliganing." );
		table.AddColumn( "situation" );
		table.AddColumn( "no mulligan" );
		table.AddColumn( "mulligan" );
		table.AddString( "screwed" );
		table.AddProbability( no_mull_land_situation_table[ 0 ] );
		table.AddProbability( mull_land_situation_table[ 0 ] );
		table.AddString( "preferred" );
		table.AddProbability( no_mull_land_situation_table[ 1 ] );
		table.AddProbability( mull_land_situation_table[ 1 ] );
		table.AddString( "flooded" );
		table.AddProbability( no_mull_land_situation_table[ 2 ] );
		table.AddProbability( mull_land_situation_table[ 2 ] );
		table.Print();
	}

	{
		const int prefered_lower_bound = 2;
		const int prefered_upper_bound = 5;

		double no_mull_land_situation_table[ 3 ];
		double mull_land_situation_table[ 3 ];
		no_mull_state.BinDistribution( no_mull_land_situation_table, std::extent< decltype( no_mull_land_situation_table ) >::value,
			[ prefered_lower_bound, prefered_upper_bound ]( const CProbabilityDistributionEntry& entry ) -> int
			{
				if ( entry.m_player.m_hand.m_lands < prefered_lower_bound ) return 0;
				if ( entry.m_player.m_hand.m_lands > prefered_upper_bound ) return 2;
				return 1;
			}
		);
		mull_state.BinDistribution( mull_land_situation_table, std::extent< decltype( mull_land_situation_table ) >::value,
			[ prefered_lower_bound, prefered_upper_bound ]( const CProbabilityDistributionEntry& entry ) -> int
			{
				if ( entry.m_player.m_hand.m_lands < prefered_lower_bound ) return 0;
				if ( entry.m_player.m_hand.m_lands > prefered_upper_bound ) return 2;
				return 1;
			}
		);
		CTable table( "Land situation distribution with and without mulliganing, counting 1 land or less as screwed." );
		table.AddColumn( "situation" );
		table.AddColumn( "no mulligan" );
		table.AddColumn( "mulligan" );
		table.AddString( "very screwed" );
		table.AddProbability( no_mull_land_situation_table[ 0 ] );
		table.AddProbability( mull_land_situation_table[ 0 ] );
		table.AddString( "satisfactory" );
		table.AddProbability( no_mull_land_situation_table[ 1 ] );
		table.AddProbability( mull_land_situation_table[ 1 ] );
		table.AddString( "flooded" );
		table.AddProbability( no_mull_land_situation_table[ 2 ] );
		table.AddProbability( mull_land_situation_table[ 2 ] );
		table.Print();

	}
	*/

	/*
	{
		CProbabilityDistribution state( 40, 17 );
		state.DrawCards( 7 );
		state.MulliganIf( MulliganRules::StandardDeck_OnTheDraw );

		CTable table( "Land count distribution after mulligans." );
		table.AddColumn( "land" );
		table.AddColumn( "probability" );
		
		for ( int land = 0; land < 8; ++land )
		{
			const double probability_of_land_count = state.TakeMeasurement(
				[ land ]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands == land;
				}
			);
			table.AddInteger( land );
			table.AddProbability( probability_of_land_count );
		}

		state.DrawCards( 4 );
		const double chance_of_competitive_start = state.TakeMeasurement( competitive_start_measurement );
		table.AddAddendum( "overall chance of a competitive start" );
		table.AddProbability( chance_of_competitive_start );

		table.Print();
	}
	*/
}
