#include "magic_probability_toolkit.h"
#include "table_output.h"
#include "mulligan_rules.h"

namespace mlr
{

void CCalculateMulliganRulesResults::PrintTable( const char* header, int min_hand_size )
{
	assert( min_hand_size >= 1 );

	tbo::CTable table( header );
	table.AddColumn( "   " );
	table.AddColumn( " 0" );
	table.AddColumn( " 1" );
	table.AddColumn( " 2" );
	table.AddColumn( " 3" );
	table.AddColumn( " 4" );
	table.AddColumn( " 5" );
	table.AddColumn( " 6" );
	table.AddColumn( " 7" );
	for ( int starting_hand_size = 7; starting_hand_size > min_hand_size - 1; --starting_hand_size )
	{
		table.AddInteger( starting_hand_size );
		for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= 7; ++lands_in_starting_hand )
		{
			if ( lands_in_starting_hand > starting_hand_size )
			{
				table.AddString( "" );
			}
			else
			{
				table.AddString( m_rules.IsMulliganAdvised( starting_hand_size, lands_in_starting_hand ) ? "X" : "-" );
			}
		}
	}
	table.Print();
}

void CCalculateMulliganRulesResults::PrintComplexTable( const char* header )
{
	tbo::CTable table( header );
	table.AddColumn( "   " );
	table.AddColumn( " 0" );
	table.AddColumn( " 1" );
	table.AddColumn( " 2" );
	table.AddColumn( " 3" );
	table.AddColumn( " 4" );
	table.AddColumn( " 5" );
	table.AddColumn( " 6" );
	table.AddColumn( " 7" );
	for ( int starting_hand_size = 7; starting_hand_size > 0; --starting_hand_size )
	{
		table.AddInteger( starting_hand_size );
		for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= 7; ++lands_in_starting_hand )
		{
			if ( lands_in_starting_hand > starting_hand_size )
			{
				table.AddString( "" );
			}
			else
			{
				static const char* const glyphs[] = { "X", "?", "-" };
				table.AddString( glyphs[ m_complex_mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] ] );
			}
		}
	}
	table.Print();
}

void CCalculateMulliganRulesResults::PrintCleanTable( const char* header )
{
	tbo::CTable table( header );
	table.AddColumn( "   " );
	table.AddColumn( "     0" );
	table.AddColumn( "     1" );
	table.AddColumn( "     2" );
	table.AddColumn( "     3" );
	table.AddColumn( "     4" );
	table.AddColumn( "     5" );
	table.AddColumn( "     6" );
	table.AddColumn( "     7" );
	for ( int starting_hand_size = 7; starting_hand_size > 4; --starting_hand_size )
	{
		table.AddInteger( starting_hand_size );
		for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= 7; ++lands_in_starting_hand )
		{
			if ( lands_in_starting_hand > starting_hand_size )
			{
				table.AddString( "     " );
			}
			else
			{
				if ( m_complex_mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] == 0 )
				{
					table.AddString( "    X" );
				}
				else if ( m_complex_mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] == 1 )
				{
					double deviation = m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ] - m_overall_chance_of_competitive_start_table[ starting_hand_size - 1 ];
					const double larger = std::max( m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ], m_overall_chance_of_competitive_start_table[ starting_hand_size - 1 ] );
					deviation /= larger;
					table.AddProbability( deviation, 0 );
				}
				else if ( m_complex_mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] == 2 )
				{
					table.AddString( "    -" );
				}
			}
		}
	}
	table.Print();
}

void CCalculateMulliganRulesResults::PrintEntireCalculation( const char* header )
{
	for ( int starting_hand_size = m_first_interesting_hand_size - 1; starting_hand_size < 8; ++starting_hand_size )
	{
		if ( m_hand_size_secondary_criteria_used_for >= starting_hand_size )
		{
			tbo::CTable table( "%s  Chance of secondary criteria with various land counts in a %d-card hand.", header, starting_hand_size );
			table.AddColumn( "land" );
			table.AddColumn( "spells" );
			table.AddColumn( "probability" );

			for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
			{
				table.AddInteger( lands_in_starting_hand );
				table.AddInteger( starting_hand_size - lands_in_starting_hand );
				table.AddProbability( m_chance_of_secondary_criteria_table[ starting_hand_size ][ lands_in_starting_hand ] );
			}

			table.AddAddendum( "mulligan-adjusted chance of secondary criteria" );
			table.AddProbability( m_overall_chance_of_secondary_criteria_table[ starting_hand_size ] );
		
			table.Print();
		}

		{
			tbo::CTable table( "%s  Chance of a competitive start with various land counts in a %d-card hand.", header, starting_hand_size );
			table.AddColumn( "land" );
			table.AddColumn( "spells" );
			table.AddColumn( "probability" );

			for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
			{
				table.AddInteger( lands_in_starting_hand );
				table.AddInteger( starting_hand_size - lands_in_starting_hand );
				table.AddProbability( m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ] );
			}
		
			table.Print();
		}

		{
			tbo::CTable comparison_table( "%s  Comparison of competitive start chance between various land counts in a %d-card hand and the average %d-card hand.", header, starting_hand_size, starting_hand_size - 1 );
			comparison_table.AddColumn( "land" );
			comparison_table.AddColumn( "prob of comp start" );
			comparison_table.AddColumn( "" );
			comparison_table.AddColumn( "%d-card hand", starting_hand_size - 1 );
			comparison_table.AddColumn( "result" );

			tbo::CTable overall_chance_table( "%s  Overall chance of a competitive start a %d-card hand with adjusted probability on mulliganed hands.", header, starting_hand_size );
			overall_chance_table.AddColumn( "land" );
			overall_chance_table.AddColumn( "prob of hand" );
			overall_chance_table.AddColumn( "" );
			overall_chance_table.AddColumn( "prob of comp start" );
			overall_chance_table.AddColumn( "" );
			overall_chance_table.AddColumn( "" );

			for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
			{
				double chance = m_rules.m_table[ starting_hand_size ][ lands_in_starting_hand ]
					? m_overall_chance_of_competitive_start_table[ starting_hand_size - 1 ]
					: m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ];

				const double result = m_chance_of_drawing_hand_table[ starting_hand_size ][ lands_in_starting_hand ] * chance;

				comparison_table.AddInteger( lands_in_starting_hand );
				comparison_table.AddProbability( m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ] );
				comparison_table.AddString( m_rules.m_table[ starting_hand_size ][ lands_in_starting_hand ] ? "<" : ">" );
				comparison_table.AddProbability( m_overall_chance_of_competitive_start_table[ starting_hand_size - 1 ] );
				comparison_table.AddString( m_rules.m_table[ starting_hand_size ][ lands_in_starting_hand ] ? "mulligan" : "keep" );

				overall_chance_table.AddInteger( lands_in_starting_hand );
				overall_chance_table.AddProbability( m_chance_of_drawing_hand_table[ starting_hand_size ][ lands_in_starting_hand ] );
				overall_chance_table.AddString( "x" );
				overall_chance_table.AddProbability( chance );
				overall_chance_table.AddString( "=" );
				overall_chance_table.AddProbability( result );
			}

			overall_chance_table.AddAddendum( "mulligan-adjusted chance of a competitive start" );
			overall_chance_table.AddProbability( m_overall_chance_of_competitive_start_table[ starting_hand_size ] );
			overall_chance_table.AddAddendum( "unadjusted chance of a competitive start" );
			overall_chance_table.AddProbability( m_unadjusted_chance_of_competitive_start[ starting_hand_size ] );

			comparison_table.Print();

			overall_chance_table.Print();
		}
	}
}

};