#ifndef TURN_COST_H
#define TURN_COST_H

#include "magic_probability_toolkit.h"
#include "table_output.h"

namespace ctc
{

template< typename TProbabilityDistributionType, typename TExecuteTurn >
double CalculateTurnCost( const typename TProbabilityDistributionType::TPileType& deck, TExecuteTurn execute_turn, mpt::CMulliganRulesSet< typename TProbabilityDistributionType::TCardSetType > rules_set )
{
	auto get_turn = []( const TProbabilityDistributionType::TProbabilityDistributionEntry& entry ) -> int
	{
		typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;
		return entry.m_player.m_hand.m_cards[ Card::Turn ];
	};

	auto set_first_turn = []( TProbabilityDistributionType::TProbabilityDistributionEntry& entry )
	{
		typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;
		entry.m_player.m_hand.m_cards[ Card::Turn ] = 1;
	};

	double average_turn_cost = 0.0;

	{
		TProbabilityDistributionType state( deck );
		state.DrawStartingHandFromRules( rules_set.m_on_the_play );

		state.ForEach( set_first_turn );

		state.TakeTurnsUntil( execute_turn );
		const double turn_cost = state.TakeMeasurement( get_turn );

		average_turn_cost += turn_cost;
	}

	{
		TProbabilityDistributionType state( deck );
		state.DrawStartingHandFromRules( rules_set.m_on_the_draw );

		state.ForEach( set_first_turn );
		state.DrawCard();

		state.TakeTurnsUntil( execute_turn );
		const double turn_cost = state.TakeMeasurement( get_turn );

		average_turn_cost += turn_cost;
	}

	average_turn_cost /= 2.0;

	return average_turn_cost;
}

template< typename TProbabilityDistributionType, typename TInitialHandEvaluation, typename TCompetitiveStartMeasurement, typename TExecuteTurn >
double CalculateTurnCost( const typename TProbabilityDistributionType::TPileType& deck, TInitialHandEvaluation initial_hand_evaluation, TCompetitiveStartMeasurement competitive_start_measurement, TExecuteTurn execute_turn )
{
	mpt::CMulliganRulesSet< TProbabilityDistributionType::TCardSetType > rules_set = CalculateMulliganRulesSet<
		TProbabilityDistributionType,
		TInitialHandEvaluation,
		TCompetitiveStartMeasurement
	>( deck, initial_hand_evaluation, competitive_start_measurement );

	return CalculateTurnCost<
		TProbabilityDistributionType,
		TExecuteTurn
	>( deck, execute_turn, rules_set );
}

}

#endif // CALCULATE_TURN_COST_H